from __future__ import annotations

import copy
import json
import logging
import operator
import os
import pickle
import shutil
import sys
import time
import warnings
from datetime import timedelta
from typing import Callable, Dict, List, Optional, Union

import lightning.pytorch as pl
import numpy as np
import pandas as pd
import torch
import yaml
from lightning.pytorch.strategies import DeepSpeedStrategy
from omegaconf import DictConfig, OmegaConf
from packaging import version
from torch import nn

from autogluon.common.utils.resource_utils import ResourceManager
from autogluon.common.utils.try_import import try_import_ray
from autogluon.core.metrics import Scorer, get_metric
from autogluon.core.utils.loaders import load_pd

from .. import version as ag_version
from ..constants import (
    BEST,
    BEST_K_MODELS_FILE,
    BINARY,
    COLUMN_FEATURES,
    DDP,
    DEEPSPEED_MIN_PL_VERSION,
    DEEPSPEED_MODULE,
    DEEPSPEED_OFFLOADING,
    DISTILLER,
    DOCUMENT,
    FEATURE_EXTRACTION,
    FEATURES,
    FEW_SHOT,
    FEW_SHOT_CLASSIFICATION,
    GREEDY_SOUP,
    IMAGE_BASE64_STR,
    IMAGE_BYTEARRAY,
    IMAGE_PATH,
    LABEL,
    LAST_CHECKPOINT,
    LOGITS,
    MASKS,
    MAX,
    MIN,
    MODEL_CHECKPOINT,
    MULTICLASS,
    NER,
    RAY_TUNE_CHECKPOINT,
    REGRESSION,
    TEXT,
    TEXT_NER,
    TORCH_COMPILE_MIN_VERSION,
    UNIFORM_SOUP,
    Y_PRED,
    Y_PRED_PROB,
    Y_TRUE,
    ZERO_SHOT_IMAGE_CLASSIFICATION,
)
from ..data import (
    BaseDataModule,
    MultiModalFeaturePreprocessor,
    create_fusion_data_processors,
    data_to_df,
    get_mixup,
    infer_column_types,
    infer_dtypes_by_model_names,
    infer_output_shape,
    infer_problem_type,
    infer_scarcity_mode_by_data_size,
    init_df_preprocessor,
    is_image_column,
    split_train_tuning_data,
    turn_on_off_feature_column_info,
)
from ..models import (
    create_fusion_model,
    get_model_postprocess_fn,
    is_lazy_weight_tensor,
    list_timm_models,
    select_model,
)
from ..optim import (
    compute_score,
    get_aug_loss_func,
    get_loss_func,
    get_minmax_mode,
    get_norm_layer_param_names,
    get_peft_param_names,
    get_stopping_threshold,
    get_torchmetric,
    infer_metrics,
)
from ..optim.lit_distiller import DistillerLitModule
from ..optim.lit_module import LitModule
from ..utils import (
    AutoMMModelCheckpoint,
    AutoMMModelCheckpointIO,
    DDPPredictionWriter,
    DistillationMixin,
    ExportMixin,
    LogFilter,
    RealtimeMixin,
    apply_log_filter,
    average_checkpoints,
    compute_inference_batch_size,
    compute_num_gpus,
    extract_from_output,
    filter_hyperparameters,
    get_config,
    get_dir_ckpt_paths,
    get_gpu_message,
    get_load_ckpt_paths,
    get_local_pretrained_config_paths,
    hyperparameter_tune,
    infer_precision,
    infer_problem_type_by_eval_metric,
    is_interactive_env,
    is_interactive_strategy,
    logits_to_prob,
    on_fit_end_message,
    on_fit_per_run_start_message,
    on_fit_start_message,
    run_ddp_only_once,
    save_pretrained_model_configs,
    setup_save_path,
    split_hyperparameters,
    tensor_to_ndarray,
    update_config_by_rules,
    update_hyperparameters,
    update_tabular_config_by_resources,
)
from ..utils.problem_types import PROBLEM_TYPES_REG

pl_logger = logging.getLogger("lightning")
pl_logger.propagate = False  # https://github.com/Lightning-AI/lightning/issues/4621
logger = logging.getLogger(__name__)


class BaseLearner(ExportMixin, DistillationMixin, RealtimeMixin):
    """ """

    def __init__(
        self,
        label: Optional[str] = None,
        problem_type: Optional[str] = None,
        presets: Optional[str] = None,
        eval_metric: Optional[Union[str, Scorer]] = None,
        hyperparameters: Optional[dict] = None,
        path: Optional[str] = None,
        verbosity: Optional[int] = 2,
        warn_if_exist: Optional[bool] = True,
        enable_progress_bar: Optional[bool] = None,
        pretrained: Optional[bool] = True,
        validation_metric: Optional[str] = None,
        **kwargs,
    ):
        """
        Parameters
        ----------
        label
            Name of the column that contains the target variable to predict.
        problem_type
            Type of the prediction problem. We support standard problems like

            - 'binary': Binary classification
            - 'multiclass': Multi-class classification
            - 'regression': Regression
            - 'classification': Classification problems include 'binary' and 'multiclass' classification.

            In addition, we support advanced problems such as

            - 'object_detection': Object detection
            - 'ner' or 'named_entity_recognition': Named entity extraction
            - 'text_similarity': Text-text similarity problem
            - 'image_similarity': Image-image similarity problem
            - 'image_text_similarity': Text-image similarity problem
            - 'feature_extraction': Extracting feature (only support inference)
            - 'zero_shot_image_classification': Zero-shot image classification (only support inference)
            - 'few_shot_classification': Few-shot classification for image or text data.

            For certain problem types, the default behavior is to load a pretrained model based on
            the presets / hyperparameters and the learner will support zero-shot inference
            (running inference without .fit()). This includes the following
            problem types:

            - 'object_detection'
            - 'text_similarity'
            - 'image_similarity'
            - 'image_text_similarity'
            - 'feature_extraction'
            - 'zero_shot_image_classification'
            - 'few_shot_classification'
        presets
            Presets regarding model quality, e.g., best_quality, high_quality, and medium_quality.
        eval_metric
            Evaluation metric name. If `eval_metric = None`, it is automatically chosen based on `problem_type`.
            Defaults to 'accuracy' for multiclass classification, `roc_auc` for binary classification, and 'root_mean_squared_error' for regression.
        hyperparameters
            This is to override some default configurations.
            For example, changing the text and image backbones can be done by formatting:

            a string
            hyperparameters = "model.hf_text.checkpoint_name=google/electra-small-discriminator model.timm_image.checkpoint_name=swin_small_patch4_window7_224"

            or a list of strings
            hyperparameters = ["model.hf_text.checkpoint_name=google/electra-small-discriminator", "model.timm_image.checkpoint_name=swin_small_patch4_window7_224"]

            or a dictionary
            hyperparameters = {
                            "model.hf_text.checkpoint_name": "google/electra-small-discriminator",
                            "model.timm_image.checkpoint_name": "swin_small_patch4_window7_224",
                        }
        path
            Path to directory where models and intermediate outputs should be saved.
            If unspecified, a time-stamped folder called "AutogluonAutoMM/ag-[TIMESTAMP]"
            will be created in the working directory to store all models.
            Note: To call `fit()` twice and save all results of each fit,
            you must specify different `path` locations or don't specify `path` at all.
            Otherwise files from first `fit()` will be overwritten by second `fit()`.
        verbosity
            Verbosity levels range from 0 to 4 and control how much information is printed.
            Higher levels correspond to more detailed print statements (you can set verbosity = 0 to suppress warnings).
            If using logging, you can alternatively control amount of information printed via `logger.setLevel(L)`,
            where `L` ranges from 0 to 50
            (Note: higher values of `L` correspond to fewer print statements, opposite of verbosity levels)
        warn_if_exist
            Whether to raise warning if the specified path already exists.
        enable_progress_bar
            Whether to show progress bar. It will be True by default and will also be
            disabled if the environment variable os.environ["AUTOMM_DISABLE_PROGRESS_BAR"] is set.
        pretrained
            Whether to init model with pretrained weights. If False, it creates a model with random initialization.
        validation_metric
            Validation metric name. If `validation_metric = None`, it is automatically chosen based on `problem_type`.
            Defaults to 'accuracy' for multiclass classification, `roc_auc` for binary classification, and 'root_mean_squared_error' for regression.
        """
        self._eval_metric_name = None
        self._eval_metric_func = None
        if isinstance(eval_metric, str):
            self._eval_metric_name = eval_metric.lower()
            self.set_eval_metric_func()
        elif isinstance(eval_metric, Scorer):
            self._eval_metric_name = eval_metric.name
            self._eval_metric_func = eval_metric

        self._problem_type = infer_problem_type_by_eval_metric(
            eval_metric_name=self._eval_metric_name,
            problem_type=problem_type,
        )

        self._label_column = label
        self._presets = presets.lower() if presets else None
        self._validation_metric_name = validation_metric.lower() if validation_metric else None
        self._minmax_mode = None
        self._output_shape = None
        self._ckpt_path = None
        self._pretrained_path = None
        self._pretrained = pretrained
        self._config = None
        self._df_preprocessor = None
        self._column_types = None
        self._data_processors = None
        self._model_postprocess_fn = None
        self._model = None
        self._resume = False
        self._verbosity = verbosity
        self._warn_if_exist = warn_if_exist
        self._enable_progress_bar = enable_progress_bar if enable_progress_bar is not None else True
        self._fit_called = False  # Flag whether is continuous training.
        self._save_path = path
        self._hyperparameters = hyperparameters
        self._advanced_hyperparameters = None
        self._hyperparameter_tune_kwargs = None
        self._train_data = None
        self._tuning_data = None
        self._is_hpo = False
        self._teacher_learner = None
        self._fit_args = None
        # Summary statistics used in fit summary. TODO: wrap it in a class.
        self._total_train_time = None
        self._best_score = None

        self._log_filters = [
            ".*does not have many workers.* in the `DataLoader` init to improve performance.*",
            "Checkpoint directory .* exists and is not empty.",
        ]

    @property
    def path(self):
        return self._save_path

    @property
    def label(self):
        return self._label_column

    @property
    def problem_type(self):
        return self._problem_type

    @property
    def problem_property(self):
        if self.problem_type is None:
            return None
        else:
            return PROBLEM_TYPES_REG.get(self.problem_type)

    @property
    def column_types(self):
        return self._column_types

    @property
    def eval_metric(self):
        return self._eval_metric_name

    @property
    def validation_metric(self):
        return self._validation_metric_name

    @property
    def total_parameters(self) -> int:
        return sum(p.numel() if not is_lazy_weight_tensor(p) else 0 for p in self._model.parameters())

    @property
    def trainable_parameters(self) -> int:
        return sum(
            p.numel() if not is_lazy_weight_tensor(p) else 0 for p in self._model.parameters() if p.requires_grad
        )

    @property
    def model_size(self) -> float:
        """
        Returns the model size in Megabyte.
        """
        model_size = sum(
            p.numel() * p.element_size() if not is_lazy_weight_tensor(p) else 0 for p in self._model.parameters()
        )
        return model_size * 1e-6  # convert to megabytes

    def set_eval_metric_func(self):
        from .matching import MatchingLearner
        from .ner import NERLearner
        from .object_detection import ObjectDetectionLearner
        from .semantic_segmentation import SemanticSegmentationLearner

        if (
            not isinstance(self, (NERLearner, SemanticSegmentationLearner, MatchingLearner, ObjectDetectionLearner))
            and self._eval_metric_func is None
        ):
            self._eval_metric_func = get_metric(self._eval_metric_name)

    def ensure_fit_ready(self):
        if self._problem_type and not self.problem_property.support_fit:
            raise RuntimeError(
                f"The problem_type='{self._problem_type}' does not support `learner.fit()`. "
                f"You may try to use `learner.predict()` or `learner.evaluate()`."
            )

    def infer_problem_type(self, train_data: pd.DataFrame):
        if self._fit_called:
            return
        if self._label_column:
            if isinstance(train_data, str):
                train_data = load_pd.load(train_data)
            self._problem_type = infer_problem_type(
                y_train_data=train_data[self._label_column],
                provided_problem_type=self._problem_type,
            )

    def setup_save_path(self, save_path: str):
        self._save_path = setup_save_path(
            resume=self._resume,
            old_save_path=self._save_path,
            proposed_save_path=save_path,
            raise_if_exist=True,
            warn_if_exist=False,
            fit_called=self._fit_called,
        )

    def infer_column_types(
        self, column_types: Optional[Dict] = None, data: Optional[pd.DataFrame] = None, is_train=True
    ):
        if is_train and self._fit_called:
            return
        elif is_train:
            self._column_types = infer_column_types(
                data=self._train_data,
                valid_data=self._tuning_data,
                label_columns=self._label_column,
                provided_column_types=column_types,
                problem_type=self._problem_type,  # used to update the corresponding column type
            )
            logger.debug(f"column_types: {self._column_types}")
            logger.debug(f"image columns: {[k for k, v in self._column_types.items() if v == 'image_path']}")
        elif column_types is None:
            allowable_dtypes, fallback_dtype = infer_dtypes_by_model_names(model_config=self._config.model)
            return infer_column_types(
                data=data,
                label_columns=self._label_column,
                problem_type=self._problem_type,
                allowable_column_types=allowable_dtypes,
                fallback_column_type=fallback_dtype,
            )
        else:
            return column_types

    def infer_output_shape(self):
        if self._fit_called:
            return
        self._output_shape = infer_output_shape(
            label_column=self._label_column,
            data=self._train_data,
            problem_type=self._problem_type,
        )

    def prepare_train_tuning_data(
        self,
        train_data: Union[pd.DataFrame, str],
        tuning_data: Optional[Union[pd.DataFrame, str]],
        holdout_frac: Optional[float],
        seed: Optional[int],
    ):
        if isinstance(train_data, str):
            train_data = load_pd.load(train_data)
        if isinstance(tuning_data, str):
            tuning_data = load_pd.load(tuning_data)

        if tuning_data is None:
            train_data, tuning_data = split_train_tuning_data(
                data=train_data,
                holdout_frac=holdout_frac,
                problem_type=self._problem_type,
                label_column=self._label_column,
                random_state=seed,
            )

        self._train_data = train_data
        self._tuning_data = tuning_data

    def detect_data_scarcity_mode(self):
        # Determine data scarcity mode, i.e. a few-shot scenario
        scarcity_mode = infer_scarcity_mode_by_data_size(
            df_train=self._train_data, scarcity_threshold=50
        )  # Add as separate hyperparameter somewhere?
        if scarcity_mode == FEW_SHOT and (
            not self._presets or FEW_SHOT not in self._presets
        ):  # TODO: check for data type
            logger.warning(
                f"Detected data scarcity. Consider running using the problem_type '{FEW_SHOT_CLASSIFICATION}' for better performance."
            )

    def update_attributes(
        self,
        config: Optional[Dict] = None,
        df_preprocessor: Optional[MultiModalFeaturePreprocessor] = None,
        data_processors: Optional[Dict] = None,
        model: Optional[nn.Module] = None,
        model_postprocess_fn: Optional[Callable] = None,
        best_score: Optional[float] = None,
        **kwargs,
    ):
        if config:
            self._config = config
        if df_preprocessor:
            self._df_preprocessor = df_preprocessor
        if data_processors:
            self._data_processors = data_processors
        if model:
            self._model = model
        if model_postprocess_fn:
            self._model_postprocess_fn = model_postprocess_fn
        if best_score:
            self._best_score = best_score

    def infer_validation_metric(self, is_matching: Optional[bool] = False):
        if self._fit_called:
            return
        self._validation_metric_name, self._eval_metric_name = infer_metrics(
            problem_type=self._problem_type,
            eval_metric=self._eval_metric_name if self._eval_metric_func is None else self._eval_metric_func,
            validation_metric_name=self._validation_metric_name,
            is_matching=is_matching,
        )
        self.set_eval_metric_func()
        self._minmax_mode = get_minmax_mode(self._validation_metric_name)
        logger.debug(f"validation_metric_name: {self._validation_metric_name}")
        logger.debug(f"minmax_mode: {self._minmax_mode}")

    def update_hyperparameters(self, hyperparameters: Dict, hyperparameter_tune_kwargs: Dict):
        problem_type = self._pipeline if hasattr(self, "_pipeline") else self._problem_type  # matching uses pipeline
        if self._hyperparameters and hyperparameters:
            self._hyperparameters.update(hyperparameters)
        elif hyperparameters:
            self._hyperparameters = hyperparameters

        if self._hyperparameter_tune_kwargs and hyperparameter_tune_kwargs:
            self._hyperparameter_tune_kwargs.update(hyperparameter_tune_kwargs)
        elif hyperparameter_tune_kwargs:
            self._hyperparameter_tune_kwargs = hyperparameter_tune_kwargs

        self._hyperparameters, self._hyperparameter_tune_kwargs = update_hyperparameters(
            problem_type=problem_type,
            presets=self._presets,
            provided_hyperparameters=self._hyperparameters,
            provided_hyperparameter_tune_kwargs=self._hyperparameter_tune_kwargs,
        )
        # split out the hyperparameters whose values are complex objects
        self._hyperparameters, self._advanced_hyperparameters = split_hyperparameters(self._hyperparameters)
        self._is_hpo = True if self._hyperparameter_tune_kwargs else False
        if self._is_hpo:
            try_import_ray()
            self._hyperparameters = filter_hyperparameters(
                hyperparameters=self._hyperparameters,
                column_types=self._column_types,
                config=self._config,
                fit_called=self._fit_called,
            )

    def fit_sanity_check(self):
        assert not self._resume or not self._is_hpo, "You can not resume training with HPO."
        if self._is_hpo and hasattr(self, "_teacher_learner") and self._teacher_learner is not None:
            assert isinstance(
                self._teacher_learner, str
            ), "HPO with distillation only supports passing a path to the learner."

    def prepare_fit_args(
        self,
        time_limit: int,
        seed: int,
        standalone: Optional[bool] = True,
        clean_ckpts: Optional[bool] = True,
    ):
        if time_limit is not None:
            time_limit = timedelta(seconds=time_limit)
        self._fit_args = dict(
            max_time=time_limit,
            save_path=self._save_path,  # In HPO mode, this would be overwritten by per trial path.
            ckpt_path=None if self._is_hpo else self._ckpt_path,
            resume=False if self._is_hpo else self._resume,
            enable_progress_bar=False if self._is_hpo else self._enable_progress_bar,
            seed=seed,
            hyperparameters=self._hyperparameters,  # In HPO mode, this would be overwritten by sampled hyperparameters.
            advanced_hyperparameters=self._advanced_hyperparameters,
            standalone=standalone,
            clean_ckpts=clean_ckpts,
        )
        if self._fit_called:  # continuous training
            continuous_train_args = dict(
                config=self._config,
                df_preprocessor=self._df_preprocessor,
                data_processors=self._data_processors,
                model=self._model,
            )
            self._fit_args.update(continuous_train_args)

    def execute_fit(self):
        if self._is_hpo:
            self._fit_args["learner"] = self
            hyperparameter_tune(
                hyperparameter_tune_kwargs=self._hyperparameter_tune_kwargs,
                resources=dict(num_gpus=ResourceManager.get_gpu_count_torch()),  # TODO: allow customizing GPUs
                **self._fit_args,
            )
            return dict()
        else:
            attributes = self.fit_per_run(**self._fit_args)
            self.update_attributes(**attributes)  # only update attributes for non-HPO mode
            return attributes

    def on_fit_start(
        self,
        presets: Optional[str] = None,
        teacher_learner: Optional[Union[str, BaseLearner]] = None,
    ):
        self.ensure_fit_ready()
        if presets:
            if self._fit_called:
                warnings.warn("Ignoring the provided `presets` as fit() was called before.", UserWarning)
            else:
                self._presets = presets
        if teacher_learner:
            self._teacher_learner = teacher_learner

        logger.info(on_fit_start_message(path=self._save_path))
        training_start = time.time()
        return training_start

    def on_fit_end(
        self,
        training_start: float,
        strategy: Optional[str] = None,
        strict_loading: Optional[bool] = True,
        standalone: Optional[bool] = True,
        clean_ckpts: Optional[bool] = True,
    ):
        self._fit_called = True
        if not self._is_hpo:
            # top_k_average is called inside hyperparameter_tune() when building the final predictor.
            self.top_k_average(
                save_path=self._save_path,
                top_k_average_method=self._config.optim.top_k_average_method,
                strategy=strategy,
                strict_loading=strict_loading,
                # Not strict loading if using parameter-efficient finetuning
                standalone=standalone,
                clean_ckpts=clean_ckpts,
            )

        training_end = time.time()
        self._total_train_time = training_end - training_start
        # TODO(?) We should have a separate "_post_training_event()" for logging messages.
        logger.info(on_fit_end_message(self._save_path))

    def fit(
        self,
        train_data: Union[pd.DataFrame, str],
        presets: Optional[str] = None,
        tuning_data: Optional[Union[pd.DataFrame, str]] = None,
        time_limit: Optional[int] = None,
        save_path: Optional[str] = None,
        hyperparameters: Optional[Union[str, Dict, List[str]]] = None,
        column_types: Optional[Dict] = None,
        holdout_frac: Optional[float] = None,
        teacher_learner: Union[str, BaseLearner] = None,
        seed: Optional[int] = 0,
        standalone: Optional[bool] = True,
        hyperparameter_tune_kwargs: Optional[Dict] = None,
        clean_ckpts: Optional[bool] = True,
        **kwargs,
    ):
        self.setup_save_path(save_path=save_path)
        training_start = self.on_fit_start(presets=presets, teacher_learner=teacher_learner)
        self.infer_problem_type(train_data=train_data)
        self.prepare_train_tuning_data(
            train_data=train_data,
            tuning_data=tuning_data,
            holdout_frac=holdout_frac,
            seed=seed,
        )
        self.infer_column_types(column_types=column_types)
        self.infer_output_shape()
        self.infer_validation_metric()
        self.update_hyperparameters(
            hyperparameters=hyperparameters,
            hyperparameter_tune_kwargs=hyperparameter_tune_kwargs,
        )
        self.fit_sanity_check()
        self.prepare_fit_args(
            time_limit=time_limit,
            seed=seed,
            standalone=standalone,
            clean_ckpts=clean_ckpts,
        )
        fit_returns = self.execute_fit()
        self.on_fit_end(
            training_start=training_start,
            strategy=fit_returns.get("strategy", None),
            strict_loading=fit_returns.get("strict_loading", True),
            standalone=standalone,
            clean_ckpts=clean_ckpts,
        )

        return self

    def init_pretrained(self):
        # split out the hyperparameters whose values are complex objects
        hyperparameters, advanced_hyperparameters = split_hyperparameters(self._hyperparameters)
        if self._config is None:
            self._config = get_config(
                problem_type=self._problem_type,
                presets=self._presets,
                overrides=hyperparameters,
            )
        if self._model is None:
            assert (
                len(self._config.model.names) == 1
            ), f"Zero shot mode only supports using one model, but detects multiple models {self._config.model.names}"
            self._model = create_fusion_model(
                config=self._config,
                pretrained=self._pretrained,
                num_classes=self._output_shape,
            )
        if self._data_processors is None:
            self._data_processors = create_fusion_data_processors(
                config=self._config,
                model=self._model,
                advanced_hyperparameters=advanced_hyperparameters,
            )
        self._config = self.update_strategy_by_env(config=self._config)

    def get_config_per_run(self, config, hyperparameters):
        config = get_config(
            problem_type=self._problem_type,
            presets=self._presets,
            config=config,
            overrides=hyperparameters,  # don't use self._hyperparameters due to HPO.
            extra=[DISTILLER] if self._teacher_learner is not None else None,
        )
        config = update_config_by_rules(
            problem_type=self._problem_type,
            config=config,
        )
        config = self.update_strategy_by_env(config=config)
        return config

    def get_df_preprocessor_per_run(
        self,
        df_preprocessor,
        data=None,
        config=None,
        column_types=None,
        is_train=True,
    ):
        if df_preprocessor is None:
            if is_train:
                df_preprocessor = init_df_preprocessor(
                    config=config,
                    column_types=self._column_types,
                    label_column=self._label_column,
                    train_df_x=self._train_data.drop(columns=self._label_column),
                    train_df_y=self._train_data[self._label_column],
                )
            else:
                df_preprocessor = init_df_preprocessor(
                    config=self._config,
                    column_types=column_types,
                    label_column=self._label_column,
                    train_df_x=data,
                    train_df_y=data[self._label_column] if self._label_column in data else None,
                )

        return df_preprocessor

    @staticmethod
    def update_config_by_data_per_run(config, df_preprocessor):
        # Avoid passing tabular data with many columns to MultiHeadAttention.
        # If models have additive_attention="auto", we enable it automatically for large tables.
        config = update_tabular_config_by_resources(
            config,
            num_numerical_columns=len(df_preprocessor.numerical_feature_names),
            num_categorical_columns=len(df_preprocessor.categorical_num_categories),
        )
        config = select_model(config=config, df_preprocessor=df_preprocessor)
        return config

    def get_model_per_run(self, model, config, df_preprocessor):
        if model is None:
            model = create_fusion_model(
                config=config,
                num_classes=self._output_shape,
                num_numerical_columns=len(df_preprocessor.numerical_feature_names),
                num_categories=df_preprocessor.categorical_num_categories,
                numerical_fill_values=df_preprocessor.numerical_fill_values,
            )
        return model

    @staticmethod
    def compile_model_per_run(config, model):
        if config.env.compile.turn_on:
            assert version.parse(torch.__version__) >= version.parse(TORCH_COMPILE_MIN_VERSION), (
                f"torch.compile requires torch version >= {TORCH_COMPILE_MIN_VERSION}, "
                f"but torch version {torch.__version__} is detected."
            )
            logger.debug("Using torch.compile() in compiling the model.")
            model = torch.compile(
                model,
                mode=config.env.compile.mode,
                dynamic=config.env.compile.dynamic,
                backend=config.env.compile.backend,
            )
        return model

    @staticmethod
    def get_peft_param_names_per_run(model, config):
        peft_param_names = None
        peft = config.optim.peft
        if peft:
            norm_param_names = get_norm_layer_param_names(model)
            peft_param_names = get_peft_param_names(
                norm_param_names,
                peft=peft,
            )
        return peft_param_names

    def get_data_processors_per_run(
        self,
        data_processors,
        config=None,
        model=None,
        advanced_hyperparameters=None,
        requires_label=None,
        is_train=True,
    ):
        if is_train:
            if data_processors is None:
                data_processors = create_fusion_data_processors(
                    config=config,
                    model=model,
                    advanced_hyperparameters=advanced_hyperparameters,
                )
                data_processors_count = {k: len(v) for k, v in data_processors.items()}
                logger.debug(f"data_processors_count: {data_processors_count}")
        else:
            # For each inference call, decouple the used data processors from the learner's attribute
            data_processors = copy.copy(data_processors)
            # For prediction data with no labels provided.
            if not requires_label:
                data_processors.pop(LABEL, None)

        return data_processors

    def get_validation_metric_per_run(self):
        validation_metric, custom_metric_func = get_torchmetric(
            metric_name=self._validation_metric_name,
            num_classes=self._output_shape,
            problem_type=self._problem_type,
        )
        return validation_metric, custom_metric_func

    def get_mixup_func_per_run(self, config):
        mixup_active, mixup_func = get_mixup(
            model_config=config.model,
            mixup_config=config.data.mixup,
            num_classes=self._output_shape,
        )
        if mixup_active and (config.env.per_gpu_batch_size == 1 or config.env.per_gpu_batch_size % 2 == 1):
            warnings.warn(
                "The mixup is done on the batch."
                "The per_gpu_batch_size should be >1 and even for reasonable operation",
                UserWarning,
            )
        return mixup_active, mixup_func

    def get_loss_func_per_run(self, config, mixup_active=None):
        loss_func = get_loss_func(
            problem_type=self._problem_type,
            mixup_active=mixup_active,
            loss_func_name=config.optim.loss_func,
            config=config.optim,
        )
        aug_loss_func = get_aug_loss_func(
            config=config.optim,
            problem_type=self._problem_type,
        )
        return loss_func, aug_loss_func

    def get_model_postprocess_fn_per_run(self, loss_func):
        model_postprocess_fn = get_model_postprocess_fn(
            problem_type=self._problem_type,
            loss_func=loss_func,
        )
        return model_postprocess_fn

    def get_datamodule_per_run(
        self,
        df_preprocessor,
        data_processors,
        per_gpu_batch_size,
        num_workers,
        predict_data=None,
        is_train=True,
    ):
        if is_train and self._teacher_learner is not None:
            df_preprocessor = [df_preprocessor, self._teacher_learner._df_preprocessor]
            data_processors = [data_processors, self._teacher_learner._data_processors]
        datamodule_kwargs = dict(
            df_preprocessor=df_preprocessor,
            data_processors=data_processors,
            per_gpu_batch_size=per_gpu_batch_size,
            num_workers=num_workers,
        )
        if is_train:
            datamodule_kwargs.update(dict(train_data=self._train_data, validate_data=self._tuning_data))
        else:
            datamodule_kwargs.update(dict(predict_data=predict_data))

        datamodule = BaseDataModule(**datamodule_kwargs)
        return datamodule

    def get_optim_kwargs_per_run(
        self,
        config,
        validation_metric,
        custom_metric_func,
        loss_func,
        aug_loss_func,
        mixup_func,
        grad_steps,
    ):
        return dict(
            optim_type=config.optim.optim_type,
            lr_choice=config.optim.lr_choice,
            lr_schedule=config.optim.lr_schedule,
            lr=config.optim.lr,
            lr_decay=config.optim.lr_decay,
            end_lr=config.optim.end_lr,
            lr_mult=config.optim.lr_mult,
            weight_decay=config.optim.weight_decay,
            warmup_steps=config.optim.warmup_steps,
            track_grad_norm=config.optim.track_grad_norm,
            validation_metric=validation_metric,
            validation_metric_name=self._validation_metric_name,
            custom_metric_func=custom_metric_func,
            loss_func=loss_func,
            mixup_fn=mixup_func,
            peft=config.optim.peft,
            mixup_off_epoch=config.data.mixup.turn_off_epoch,
            skip_final_val=config.optim.skip_final_val,
            cross_modal_align=config.optim.cross_modal_align,
            cross_modal_align_weight=config.optim.cross_modal_align_weight,
            automatic_optimization=config.optim.automatic_optimization,
            accumulate_grad_batches=grad_steps,
            gradient_clip_val=config.optim.gradient_clip_val,
            gradient_clip_algorithm=config.optim.gradient_clip_algorithm,
            use_aug_optim=config.optim.lemda.turn_on,
            aug_loss_func=aug_loss_func,
            aug_lr=config.optim.lemda.lr,
            aug_weight_decay=config.optim.lemda.weight_decay,
            aug_optim_type=config.optim.lemda.optim_type,
        )

    def get_litmodule_per_run(
        self,
        model=None,
        model_postprocess_fn=None,
        peft_param_names=None,
        optim_kwargs=dict(),
        distillation_kwargs=dict(),
        is_train=True,
    ):
        if is_train:
            if self._teacher_learner is not None:
                return DistillerLitModule(
                    student_model=model,
                    teacher_model=self._teacher_learner._model,
                    **optim_kwargs,
                    **distillation_kwargs,
                )
            else:
                return LitModule(
                    model=model,
                    model_postprocess_fn=model_postprocess_fn,
                    trainable_param_names=peft_param_names,
                    **optim_kwargs,
                )
        else:
            return LitModule(
                model=self._model,
                model_postprocess_fn=self._model_postprocess_fn,
                **optim_kwargs,
            )

    def get_callbacks_per_run(self, save_path=None, config=None, litmodule=None, pred_writer=None, is_train=True):
        if not is_train:
            if pred_writer is not None:
                callbacks = [pred_writer]
            else:
                callbacks = []
            return callbacks

        checkpoint_callback = AutoMMModelCheckpoint(
            dirpath=save_path,
            save_top_k=config.optim.top_k,
            verbose=True,
            monitor=litmodule.validation_metric_name,
            mode=self._minmax_mode,
            save_last=True,
        )
        early_stopping_callback = pl.callbacks.EarlyStopping(
            monitor=litmodule.validation_metric_name,
            patience=config.optim.patience,
            mode=self._minmax_mode,
            stopping_threshold=get_stopping_threshold(self._validation_metric_name),
        )
        lr_callback = pl.callbacks.LearningRateMonitor(logging_interval="step")
        model_summary = pl.callbacks.ModelSummary(max_depth=1)
        callbacks = [
            checkpoint_callback,
            early_stopping_callback,
            lr_callback,
            model_summary,
        ]
        if self._is_hpo:
            from ..utils.hpo import get_ray_tune_ckpt_callback

            TuneReportCheckpointCallback = get_ray_tune_ckpt_callback()
            tune_report_callback = TuneReportCheckpointCallback(
                {f"{litmodule.validation_metric_name}": f"{litmodule.validation_metric_name}"},
                filename=RAY_TUNE_CHECKPOINT,
            )
            callbacks = [
                tune_report_callback,
                early_stopping_callback,
                lr_callback,
                model_summary,
            ]

        return callbacks

    @staticmethod
    def get_plugins_per_run(model, peft_param_names=None):
        custom_checkpoint_plugin = AutoMMModelCheckpointIO(
            trainable_param_names=peft_param_names,
            model_name_to_id=model.name_to_id,
        )
        return [custom_checkpoint_plugin]

    @staticmethod
    def get_tb_logger(save_path):
        return pl.loggers.TensorBoardLogger(
            save_dir=save_path,
            name="",
            version="",
        )

    @staticmethod
    def log_gpu_info(num_gpus, config):
        logger.info(
            get_gpu_message(
                detected_num_gpus=ResourceManager.get_gpu_count_torch(),
                used_num_gpus=num_gpus,
                strategy=config.env.strategy,
            )
        )

    @staticmethod
    def get_grad_steps(num_gpus, config):
        if num_gpus == 0:  # CPU only training
            grad_steps = max(
                config.env.batch_size // (config.env.per_gpu_batch_size * config.env.num_nodes),
                1,
            )
        else:
            grad_steps = max(
                config.env.batch_size // (config.env.per_gpu_batch_size * num_gpus * config.env.num_nodes),
                1,
            )
        return grad_steps

    @staticmethod
    def get_strategy_per_run(num_gpus, config):
        if (
            config.env.strategy == DEEPSPEED_OFFLOADING and num_gpus == 1 and DEEPSPEED_MODULE not in sys.modules
        ):  # Offloading currently only tested for single GPU
            assert (
                version.parse(pl.__version__) >= version.parse(DEEPSPEED_MIN_PL_VERSION)
            ), f"For DeepSpeed Offloading to work reliably you need at least lightning version {DEEPSPEED_MIN_PL_VERSION}, however, found {pl.__version__}. Please update your lightning version."
            from ..optim.deepspeed import CustomDeepSpeedStrategy

            strategy = CustomDeepSpeedStrategy(
                stage=3,
                offload_optimizer=True,
                offload_parameters=False,
                allgather_bucket_size=config.env.deepspeed_allgather_size,
                reduce_bucket_size=config.env.deepspeed_allreduce_size,
            )
        elif num_gpus <= 1:
            strategy = "auto"
        else:
            strategy = config.env.strategy
        return strategy

    def update_strategy_and_num_gpus_for_hpo(self, strategy, num_gpus):
        if self._is_hpo:
            strategy = "auto"
            num_gpus = min(num_gpus, 1)  # Currently only support one trial using one gpu.
        return strategy, num_gpus

    @staticmethod
    def get_precision_per_run(num_gpus: int, precision: Union[str, int], cpu_only_warning: Optional[bool] = True):
        return infer_precision(num_gpus=num_gpus, precision=precision, cpu_only_warning=cpu_only_warning)

    def get_num_gpus_and_strategy_per_run(
        self,
        config: Optional[DictConfig] = None,
        predict_data: Optional[pd.DataFrame] = None,
        is_train: Optional[bool] = True,
    ):
        if is_train:
            data = self._train_data
        else:
            data = predict_data
            config = self._config

        num_gpus = compute_num_gpus(
            config_num_gpus=config.env.num_gpus,
            accelerator=config.env.accelerator,
        )
        num_gpus = self.update_num_gpus_by_data_size(num_gpus=num_gpus, data=data)
        strategy = self.get_strategy_per_run(num_gpus=num_gpus, config=config)
        strategy, num_gpus = self.update_strategy_and_num_gpus_for_hpo(strategy=strategy, num_gpus=num_gpus)
        num_gpus, strategy = run_ddp_only_once(num_gpus=num_gpus, strategy=strategy)

        if is_train:
            self.log_gpu_info(num_gpus=num_gpus, config=config)

        return num_gpus, strategy

    @staticmethod
    def post_update_config_per_run(config, num_gpus, precision, strategy):
        config.env.num_gpus = num_gpus
        config.env.precision = precision
        # for deepspeed offloading, the strategy becomes is a customized strategy object instead of a string,
        # but config still needs a string.
        config.env.strategy = strategy if not config.env.strategy == DEEPSPEED_OFFLOADING else DEEPSPEED_OFFLOADING
        return config

    def init_trainer_per_run(
        self,
        num_gpus,
        precision,
        strategy,
        callbacks,
        max_time=None,
        config=None,
        tb_logger=None,
        grad_steps=None,
        plugins=None,
        enable_progress_bar=None,
        barebones=False,
        is_train=True,
    ):
        if is_train:
            trainer_kwargs = dict(
                accelerator="gpu" if num_gpus > 0 else config.env.accelerator,
                devices=num_gpus if num_gpus > 0 else "auto",
                num_nodes=config.env.num_nodes,
                precision=precision,
                strategy=strategy if strategy else "auto",
                benchmark=False,
                deterministic=config.env.deterministic,
                max_epochs=config.optim.max_epochs,
                max_steps=config.optim.max_steps,
                max_time=max_time,
                callbacks=callbacks,
                logger=tb_logger,
                log_every_n_steps=config.optim.log_every_n_steps,
                enable_progress_bar=enable_progress_bar,
                fast_dev_run=config.env.fast_dev_run,
                val_check_interval=config.optim.val_check_interval,
                check_val_every_n_epoch=config.optim.check_val_every_n_epoch,
                plugins=plugins,
            )
            if config.optim.automatic_optimization:
                trainer_kwargs.update(
                    dict(
                        gradient_clip_val=config.optim.gradient_clip_val,
                        gradient_clip_algorithm=config.optim.gradient_clip_algorithm,
                        accumulate_grad_batches=grad_steps,
                    )
                )
            blacklist_msgs = ["already configured with model summary"]
            log_filter = LogFilter(blacklist_msgs)
            with apply_log_filter(log_filter):
                trainer = pl.Trainer(**trainer_kwargs)
        else:
            blacklist_msgs = []
            if self._verbosity <= 3:  # turn off logging in prediction
                blacklist_msgs.append("Automatic Mixed Precision")
                blacklist_msgs.append("GPU available")
                blacklist_msgs.append("TPU available")
                blacklist_msgs.append("IPU available")
                blacklist_msgs.append("HPU available")
                blacklist_msgs.append("select gpus")
                blacklist_msgs.append("Trainer(barebones=True)")
            log_filter = LogFilter(blacklist_msgs)

            with apply_log_filter(log_filter):
                trainer = pl.Trainer(
                    accelerator="gpu" if num_gpus > 0 else self._config.env.accelerator,
                    devices=num_gpus if num_gpus > 0 else "auto",
                    num_nodes=self._config.env.num_nodes,
                    precision=precision,
                    strategy=strategy,
                    benchmark=False,
                    enable_progress_bar=False if barebones else self._enable_progress_bar,
                    deterministic=self._config.env.deterministic,
                    max_epochs=-1,  # Add max_epochs to disable warning
                    logger=False,
                    callbacks=callbacks,
                    barebones=barebones,
                )

        return trainer

    def run_trainer(
        self,
        trainer,
        litmodule,
        datamodule,
        ckpt_path=None,
        resume=None,
        pred_writer=None,
        is_train=True,
    ):
        with warnings.catch_warnings():
            for filter in self._log_filters:
                warnings.filterwarnings("ignore", filter)
            if is_train:
                trainer.fit(
                    litmodule,
                    datamodule=datamodule,
                    ckpt_path=ckpt_path if resume else None,  # this is to resume training that was broken accidentally
                )
            else:
                blacklist_msgs = []
                if self._verbosity <= 3:  # turn off logging in prediction
                    blacklist_msgs.append("LOCAL_RANK")
                    blacklist_msgs.append("Trainer(barebones=True)")
                log_filter = LogFilter(blacklist_msgs)
                with apply_log_filter(log_filter):
                    outputs = trainer.predict(
                        litmodule,
                        datamodule=datamodule,
                        return_predictions=pred_writer is None,
                    )
                return outputs

    def on_fit_per_run_start(self, seed, save_path):
        # TODO(?) We should have a separate "_pre_training_event()" for logging messages.
        logger.info(on_fit_per_run_start_message(save_path, self._validation_metric_name))
        pl.seed_everything(seed, workers=True)

    def on_fit_per_run_end(
        self,
        trainer: pl.Trainer,
        config: DictConfig,
        model: nn.Module,
        df_preprocessor: MultiModalFeaturePreprocessor,
        data_processors: Dict,
        save_path: str,
        standalone: bool,
    ):
        self.clean_trainer_processes(trainer=trainer, is_train=True)
        self.save(
            path=save_path,
            standalone=standalone,
            config=config,
            model=model,
            df_preprocessor=df_preprocessor,
            data_processors=data_processors,
            fit_called=True,  # fit is called on one run.
            save_model=False,  # The final model will be saved in top_k_average
        )

    def fit_per_run(
        self,
        max_time: timedelta,
        save_path: str,
        ckpt_path: str,
        resume: bool,
        enable_progress_bar: bool,
        seed: int,
        hyperparameters: Optional[Union[str, Dict, List[str]]] = None,
        advanced_hyperparameters: Optional[Dict] = None,
        config: Optional[Dict] = None,
        df_preprocessor: Optional[MultiModalFeaturePreprocessor] = None,
        data_processors: Optional[Dict] = None,
        model: Optional[nn.Module] = None,
        standalone: bool = True,
        clean_ckpts: bool = True,
    ):
        self.on_fit_per_run_start(seed=seed, save_path=save_path)
        config = self.get_config_per_run(config=config, hyperparameters=hyperparameters)
        df_preprocessor = self.get_df_preprocessor_per_run(
            df_preprocessor=df_preprocessor,
            config=config,
        )
        config = self.update_config_by_data_per_run(config=config, df_preprocessor=df_preprocessor)
        model = self.get_model_per_run(model=model, config=config, df_preprocessor=df_preprocessor)
        model = self.compile_model_per_run(config=config, model=model)
        peft_param_names = self.get_peft_param_names_per_run(model=model, config=config)
        data_processors = self.get_data_processors_per_run(
            data_processors=data_processors,
            config=config,
            model=model,
            advanced_hyperparameters=advanced_hyperparameters,
        )
        validation_metric, custom_metric_func = self.get_validation_metric_per_run()
        mixup_active, mixup_func = self.get_mixup_func_per_run(config=config)
        loss_func, aug_loss_func = self.get_loss_func_per_run(config=config, mixup_active=mixup_active)
        model_postprocess_fn = self.get_model_postprocess_fn_per_run(loss_func=loss_func)
        num_gpus, strategy = self.get_num_gpus_and_strategy_per_run(config=config)
        precision = self.get_precision_per_run(num_gpus=num_gpus, precision=config.env.precision)
        grad_steps = self.get_grad_steps(num_gpus=num_gpus, config=config)

        if max_time == timedelta(seconds=0):
            return dict(
                config=config,
                df_preprocessor=df_preprocessor,
                data_processors=data_processors,
                model=model,
                model_postprocess_fn=model_postprocess_fn,
                strict_loading=not peft_param_names,
            )
        # setup distillation in each fit_per_run call to support distillation + HPO
        distillation_kwargs = self.setup_distillation(
            model=model,
            loss_func=loss_func,
            config=config,
            data_processors=data_processors,
        )
        datamodule = self.get_datamodule_per_run(
            df_preprocessor=df_preprocessor,
            data_processors=data_processors,
            per_gpu_batch_size=config.env.per_gpu_batch_size,
            num_workers=config.env.num_workers,
        )
        optim_kwargs = self.get_optim_kwargs_per_run(
            config=config,
            validation_metric=validation_metric,
            custom_metric_func=custom_metric_func,
            loss_func=loss_func,
            aug_loss_func=aug_loss_func,
            mixup_func=mixup_func,
            grad_steps=grad_steps,
        )
        litmodule = self.get_litmodule_per_run(
            model=model,
            model_postprocess_fn=model_postprocess_fn,
            peft_param_names=peft_param_names,
            optim_kwargs=optim_kwargs,
            distillation_kwargs=distillation_kwargs,
        )
        callbacks = self.get_callbacks_per_run(save_path=save_path, config=config, litmodule=litmodule)
        plugins = self.get_plugins_per_run(model=model, peft_param_names=peft_param_names)
        tb_logger = self.get_tb_logger(save_path=save_path)
        config = self.post_update_config_per_run(
            config=config,
            num_gpus=num_gpus,
            precision=precision,
            strategy=strategy,
        )
        trainer = self.init_trainer_per_run(
            num_gpus=num_gpus,
            config=config,
            precision=precision,
            strategy=strategy,
            max_time=max_time,
            callbacks=callbacks,
            tb_logger=tb_logger,
            grad_steps=grad_steps,
            plugins=plugins,
            enable_progress_bar=enable_progress_bar,
        )

        self.run_trainer(
            trainer=trainer,
            litmodule=litmodule,
            datamodule=datamodule,
            ckpt_path=ckpt_path,
            resume=resume,
        )
        self.on_fit_per_run_end(
            save_path=save_path,
            standalone=standalone,
            trainer=trainer,
            config=config,
            df_preprocessor=df_preprocessor,
            data_processors=data_processors,
            model=model,
        )

        best_score = (
            trainer.callback_metrics[f"val_{self._validation_metric_name}"].item()
            if f"val_{self._validation_metric_name}" in trainer.callback_metrics
            else self._best_score
        )  # https://github.com/autogluon/autogluon/issues/4428

        return dict(
            config=config,
            df_preprocessor=df_preprocessor,
            data_processors=data_processors,
            model=model,
            model_postprocess_fn=model_postprocess_fn,
            best_score=best_score,
            strategy=strategy,
            strict_loading=not peft_param_names,
        )

    def top_k_average(
        self,
        save_path,
        top_k_average_method,
        strategy=None,
        last_ckpt_path=None,
        strict_loading=True,
        standalone=True,
        clean_ckpts=True,
    ):
        eval_metric = self._eval_metric_name if self._eval_metric_func is None else self._eval_metric_func
        minmax_mode = get_minmax_mode(eval_metric)
        best_k_models_yaml_path = os.path.join(save_path, BEST_K_MODELS_FILE)
        if os.path.exists(best_k_models_yaml_path):
            with open(best_k_models_yaml_path, "r") as f:
                best_k_models = yaml.safe_load(f)
        else:
            # In some cases, the training ends up too early (e.g., due to time_limit) so that there is
            # no saved best_k model checkpoints. In that scenario, we won't perform any model averaging.
            best_k_models = None
        if last_ckpt_path is None:
            last_ckpt_path = os.path.join(save_path, LAST_CHECKPOINT)

        if self._teacher_learner is not None:
            prefix = "student_model."
        else:
            prefix = "model."

        if best_k_models:
            if top_k_average_method == UNIFORM_SOUP:
                logger.info(f"Start to fuse {len(best_k_models)} checkpoints via the uniform soup algorithm.")
                ingredients = top_k_model_paths = list(best_k_models.keys())
            else:
                top_k_model_paths = [
                    v[0]
                    for v in sorted(
                        list(best_k_models.items()),
                        key=lambda ele: ele[1],
                        reverse=(minmax_mode == MAX),
                    )
                ]
                if top_k_average_method == GREEDY_SOUP:
                    # Select the ingredients based on the methods proposed in paper
                    #  "Model soups: averaging weights of multiple fine-tuned models improves accuracy without
                    #  increasing inference time", https://arxiv.org/pdf/2203.05482.pdf
                    monitor_op = {MIN: operator.le, MAX: operator.ge}[minmax_mode]
                    ingredients = [top_k_model_paths[0]]
                    if len(top_k_model_paths) > 1:
                        logger.info(
                            f"Start to fuse {len(top_k_model_paths)} checkpoints via the greedy soup algorithm."
                        )

                        self._load_state_dict(
                            path=top_k_model_paths[0],
                            prefix=prefix,
                            strict=strict_loading,
                        )
                        best_score = self.evaluate(self._tuning_data, metrics=[eval_metric])
                        best_score = next(iter(best_score.values()))
                        for i in range(1, len(top_k_model_paths)):
                            cand_avg_state_dict = average_checkpoints(
                                checkpoint_paths=ingredients + [top_k_model_paths[i]],
                            )
                            self._load_state_dict(
                                state_dict=cand_avg_state_dict,
                                prefix=prefix,
                                strict=strict_loading,
                            )
                            cand_score = self.evaluate(self._tuning_data, metrics=[eval_metric])
                            cand_score = next(iter(cand_score.values()))
                            if monitor_op(cand_score, best_score):
                                # Add new ingredient
                                ingredients.append(top_k_model_paths[i])
                                best_score = cand_score
                elif top_k_average_method == BEST:
                    ingredients = [top_k_model_paths[0]]
                else:
                    raise ValueError(
                        f"The key for 'optim.top_k_average_method' is not supported. "
                        f"We only support '{GREEDY_SOUP}', '{UNIFORM_SOUP}' and '{BEST}'. "
                        f"The provided value is '{top_k_average_method}'."
                    )
        else:
            # best_k_models is empty so we will manually save a checkpoint from the trainer
            # and use it as the main ingredients
            ingredients = [last_ckpt_path]
            top_k_model_paths = []
            # no checkpoints are available, do nothing
            if not os.path.isfile(last_ckpt_path):
                return

        # Average all the ingredients
        avg_state_dict = average_checkpoints(
            checkpoint_paths=ingredients,
        )
        self._load_state_dict(
            state_dict=avg_state_dict,
            prefix=prefix,
            strict=strict_loading,
        )

        if self._teacher_learner is not None:
            avg_state_dict = self._replace_model_name_prefix(
                state_dict=avg_state_dict,
                old_prefix="student_model",
                new_prefix="model",
            )

        if not standalone:
            checkpoint = {"state_dict": avg_state_dict}
        else:
            if isinstance(strategy, DeepSpeedStrategy):
                checkpoint = {
                    "state_dict": {
                        name.partition("module.")[2]: param
                        for name, param in strategy.model._zero3_consolidated_16bit_state_dict().items()
                    }
                }
            else:
                checkpoint = {
                    "state_dict": {"model." + name: param for name, param in self._model.state_dict().items()}
                }

        torch.save(checkpoint, os.path.join(save_path, MODEL_CHECKPOINT))  # nosec B614

        if clean_ckpts:
            # clean old checkpoints + the intermediate files stored
            for per_path in top_k_model_paths:
                if os.path.isfile(per_path):
                    os.remove(per_path)
            # remove the yaml file after cleaning the checkpoints
            if os.path.isfile(best_k_models_yaml_path):
                os.remove(best_k_models_yaml_path)
            # clean the last checkpoint
            if os.path.isfile(last_ckpt_path):
                os.remove(last_ckpt_path)

    def prepare_deepspeed_offloading(self, strategy):
        # TODO: Using optimiation_kwargs for inference is confusing and bad design. Remove as soon as fixed in lightning.
        if self._config.env.strategy == DEEPSPEED_OFFLOADING and DEEPSPEED_MODULE not in sys.modules:
            # Need to initialize DeepSpeed and optimizer as currently required in lightning's integration of deepspeed.
            from ..optim.deepspeed import CustomDeepSpeedStrategy

            strategy = CustomDeepSpeedStrategy(
                stage=3,
                offload_optimizer=True,
                offload_parameters=False,
                allgather_bucket_size=self._config.env.deepspeed_allgather_size,
                reduce_bucket_size=self._config.env.deepspeed_allreduce_size,
            )

            optim_kwargs = dict(
                optim_type=self._config.optim.optim_type,
                lr_choice=self._config.optim.lr_choice,
                lr_schedule=self._config.optim.lr_schedule,
                lr=self._config.optim.lr,
                lr_decay=self._config.optim.lr_decay,
                end_lr=self._config.optim.end_lr,
                lr_mult=self._config.optim.lr_mult,
                weight_decay=self._config.optim.weight_decay,
                warmup_steps=self._config.optim.warmup_steps,
            )
        else:
            optim_kwargs = {}

        return strategy, optim_kwargs

    def get_pred_writer(self, strategy):
        pred_writer = None
        if isinstance(strategy, str) and DDP in strategy:
            pred_writer = DDPPredictionWriter(output_dir=self._save_path, write_interval="epoch", strategy=strategy)
        return pred_writer

    @staticmethod
    def collect_predictions(outputs, trainer, pred_writer, num_gpus):
        if pred_writer is not None:
            if trainer.global_rank == 0:
                outputs = pred_writer.collect_all_gpu_results(num_gpus=num_gpus)

        return outputs

    @staticmethod
    def clean_trainer_processes(trainer, is_train=True):
        if is_train:
            msg = "Training finished,"
        else:
            msg = "Prediction finished,"
        if trainer.global_rank != 0:
            sys.exit(f"{msg} exit the process with global_rank={trainer.global_rank}...")

    def update_image_column_types(self, data):
        column_types = self._column_types
        column_types_copy = copy.deepcopy(column_types)
        for col_name, col_type in column_types.items():
            if col_type in [IMAGE_BYTEARRAY, IMAGE_PATH, IMAGE_BASE64_STR]:
                if is_image_column(data=data[col_name], col_name=col_name, image_type=IMAGE_PATH):
                    image_type = IMAGE_PATH
                elif is_image_column(
                    data=data[col_name],
                    col_name=col_name,
                    image_type=IMAGE_BYTEARRAY,
                ):
                    image_type = IMAGE_BYTEARRAY
                elif is_image_column(data=data[col_name], col_name=col_name, image_type=IMAGE_BASE64_STR):
                    image_type = IMAGE_BASE64_STR
                else:
                    image_type = col_type
                if col_type != image_type:
                    column_types_copy[col_name] = image_type
        return column_types_copy

    def data_to_df(self, data):
        if self._fit_called:
            column_names = list(self._column_types.keys())
            # remove label column since it's not required in inference.
            column_names.remove(self._label_column)
            data = data_to_df(
                data=data,
                required_columns=self._df_preprocessor.required_feature_names,
                all_columns=column_names,
            )
        else:
            data = data_to_df(data=data)

        return data

    @staticmethod
    def update_realtime_for_interactive_env(realtime: bool, num_gpus: int, barebones: bool, strategy: str):
        # TODO: support realtime inference for notebook with multi-gpus
        # realtime can initialize CUDA, which can cause failures when calling fit again in the interactive env.
        if is_interactive_strategy(strategy) and realtime:
            realtime = False
            num_gpus = min(1, num_gpus)
            barebones = True

        return realtime, num_gpus, barebones

    @staticmethod
    def update_num_gpus_by_data_size(
        num_gpus: int,
        data: pd.DataFrame,
    ):
        data_size = len(data)
        if data_size < num_gpus:
            num_gpus = data_size
        return num_gpus

    def realtime_predict(
        self,
        data: pd.DataFrame,
        df_preprocessor: Union[MultiModalFeaturePreprocessor, List[MultiModalFeaturePreprocessor]],
        data_processors: Union[Dict, List[Dict]],
        num_gpus: int,
        precision: Union[int, str],
    ) -> List[Dict]:
        """
        Perform realtime inference.

        Parameters
        ----------
        data
            A dataframe.
        df_preprocessor
            Dataframe preprocessors.
        data_processors
            Data processors.
        num_gpus
            Number of GPUs.
        precision
            The precision used in inference.

        Returns
        -------
        A list of output dicts.
        """

        batch = self.process_batch(
            data=data,
            df_preprocessor=df_preprocessor,
            data_processors=data_processors,
        )
        output = self.predict_batch(
            batch=batch,
            model=self._model,
            precision=precision,
            num_gpus=num_gpus,
        )
        return [output]

    def on_predict_per_run_start(self, data: Union[str, pd.DataFrame]):
        data = self.data_to_df(data=data)
        return data

    def get_predict_batch_size_per_run(self, num_gpus: int, strategy: str):
        return compute_inference_batch_size(
            per_gpu_batch_size=self._config.env.per_gpu_batch_size,
            inference_batch_size_ratio=self._config.env.inference_batch_size_ratio,
            num_gpus=num_gpus,
            strategy=strategy,
        )

    def on_predict_per_run_end(self, trainer):
        self.clean_trainer_processes(trainer=trainer, is_train=False)

    def predict_per_run(
        self,
        data: Union[pd.DataFrame, dict, list],
        realtime: Optional[bool],
        requires_label: Optional[bool] = False,
        barebones: Optional[bool] = False,
    ) -> List[Dict]:
        """
        Perform inference for learner.

        Parameters
        ----------
        data
            The data for inference.
        realtime
            Whether use realtime inference.
        requires_label
            Whether uses label during inference.
        barebones
            Whether to run in “barebones mode”, where all lightning's features that may impact raw speed are disabled.

        Returns
        -------
        A list of output dicts.
        """
        data = self.on_predict_per_run_start(data=data)
        column_types = self.infer_column_types(
            column_types=self._column_types,
            data=data,
            is_train=False,
        )
        df_preprocessor = self.get_df_preprocessor_per_run(
            df_preprocessor=self._df_preprocessor,
            data=data,
            column_types=column_types,
            is_train=False,
        )
        if self._fit_called:
            df_preprocessor._column_types = self.update_image_column_types(data=data)
        data_processors = self.get_data_processors_per_run(
            data_processors=self._data_processors,
            requires_label=requires_label,
            is_train=False,
        )
        num_gpus, strategy = self.get_num_gpus_and_strategy_per_run(
            predict_data=data,
            is_train=False,
        )
        precision = self.get_precision_per_run(
            num_gpus=num_gpus,
            precision=self._config.env.precision,
            cpu_only_warning=False,
        )
        batch_size = self.get_predict_batch_size_per_run(num_gpus=num_gpus, strategy=strategy)
        realtime = self.use_realtime(
            realtime=realtime,
            data=data,
            data_processors=data_processors,
            batch_size=batch_size,
        )
        realtime, num_gpus, barebones = self.update_realtime_for_interactive_env(
            realtime=realtime,
            num_gpus=num_gpus,
            barebones=barebones,
            strategy=strategy,
        )

        if realtime:
            outputs = self.realtime_predict(
                data=data,
                df_preprocessor=df_preprocessor,
                data_processors=data_processors,
                num_gpus=num_gpus,
                precision=precision,
            )
            return outputs

        strategy, optim_kwargs = self.prepare_deepspeed_offloading(strategy=strategy)
        datamodule = self.get_datamodule_per_run(
            df_preprocessor=df_preprocessor,
            data_processors=data_processors,
            per_gpu_batch_size=batch_size,
            num_workers=self._config.env.num_workers_inference,
            predict_data=data,
            is_train=False,
        )
        pred_writer = self.get_pred_writer(strategy=strategy)
        callbacks = self.get_callbacks_per_run(pred_writer=pred_writer, is_train=False)
        # TODO: remove optim_kwargs from inference
        litmodule = self.get_litmodule_per_run(optim_kwargs=optim_kwargs, is_train=False)
        trainer = self.init_trainer_per_run(
            num_gpus=num_gpus,
            precision=precision,
            strategy=strategy,
            callbacks=callbacks,
            barebones=barebones,
            is_train=False,
        )
        outputs = self.run_trainer(
            trainer=trainer,
            litmodule=litmodule,
            datamodule=datamodule,
            pred_writer=pred_writer,
            is_train=False,
        )
        outputs = self.collect_predictions(
            outputs=outputs,
            trainer=trainer,
            pred_writer=pred_writer,
            num_gpus=num_gpus,
        )
        self.on_predict_per_run_end(trainer=trainer)

        return outputs

    def ensure_predict_ready(self):
        if not self._fit_called:
            if not self._problem_type or not self.problem_property.support_zero_shot:
                raise RuntimeError(
                    f"problem_type='{self._problem_type}' does not support running inference directly. "
                    f"You need to call `learner.fit()`, or load a learner first before "
                    f"running `learner.predict()`, `learner.evaluate()` or `learner.extract_embedding()`."
                )
            else:
                self.init_pretrained()

    def on_predict_start(self):
        self.ensure_predict_ready()

    def evaluate(
        self,
        data: Union[pd.DataFrame, dict, list, str],
        metrics: Optional[Union[str, List[str]]] = None,
        return_pred: Optional[bool] = False,
        realtime: Optional[bool] = False,
        **kwargs,
    ):
        """
        Evaluate model on a test dataset.

        Parameters
        ----------
        data
            A dataframe, containing the same columns as the training data.
        metrics
            A list of metric names to report.
            If None, we only return the score for the stored `_eval_metric_name`.
        return_pred
            Whether to return the prediction result of each row.
        realtime
            Whether to do realtime inference, which is efficient for small data (default False).
            If provided None, we would infer it on based on the data modalities
            and sample number.

        Returns
        -------
        A dictionary with the metric names and their corresponding scores.
        Optionally return a dataframe of prediction results.
        """
        self.on_predict_start()
        ret_type = LOGITS
        outputs = self.predict_per_run(
            data=data,
            realtime=realtime,
            requires_label=True,
        )
        logits = extract_from_output(ret_type=ret_type, outputs=outputs)
        metric_data = {}
        if self._problem_type in [BINARY, MULTICLASS]:
            y_pred_prob = logits_to_prob(logits)
            metric_data[Y_PRED_PROB] = y_pred_prob

        y_pred = self._df_preprocessor.transform_prediction(
            y_pred=logits,
            inverse_categorical=False,
        )
        y_pred_inv = self._df_preprocessor.transform_prediction(
            y_pred=logits,
            inverse_categorical=True,
        )
        y_true = self._df_preprocessor.transform_label_for_metric(df=data)
        metric_data.update(
            {
                Y_PRED: y_pred,
                Y_TRUE: y_true,
            }
        )
        if metrics is None:
            if self._eval_metric_func:
                metrics = [self._eval_metric_func]
            else:
                metrics = [self._eval_metric_name]
        if isinstance(metrics, str) or isinstance(metrics, Scorer):
            metrics = [metrics]

        results = {}
        for per_metric in metrics:
            score = compute_score(
                metric_data=metric_data,
                metric=per_metric.lower() if isinstance(per_metric, str) else per_metric,
            )
            per_metric_name = per_metric if isinstance(per_metric, str) else per_metric.name
            results[per_metric_name] = score

        if return_pred:
            return results, self._as_pandas(data=data, to_be_converted=y_pred_inv)
        else:
            return results

    def _match_queries_and_candidates(
        self,
        query_data: Union[pd.DataFrame, dict, list],
        candidate_data: Union[pd.DataFrame, dict, list],
        return_prob: Optional[bool] = False,
    ):
        query_embeddings = self.extract_embedding(query_data, as_tensor=True)
        assert (
            len(query_embeddings) == 1
        ), f"Multiple embedding types `{query_embeddings.keys()}` exist in query data. Please reduce them to one type."
        query_embeddings = list(query_embeddings.values())[0]

        candidate_embeddings = self.extract_embedding(candidate_data, as_tensor=True)
        assert (
            len(candidate_embeddings) == 1
        ), f"Multiple embedding types `{candidate_embeddings.keys()}` exist in candidate data. Please reduce them to one type."
        candidate_embeddings = list(candidate_embeddings.values())[0]

        if return_prob:
            ret = (100.0 * query_embeddings @ candidate_embeddings.T).float().softmax(dim=-1)
        else:
            ret = (query_embeddings @ candidate_embeddings.T).argmax(dim=-1)

        ret = tensor_to_ndarray(ret)

        return ret

    def predict(
        self,
        data: Union[pd.DataFrame, dict, list, str],
        candidate_data: Optional[Union[pd.DataFrame, dict, list]] = None,
        as_pandas: Optional[bool] = None,
        realtime: Optional[bool] = False,
        **kwargs,
    ):
        """
        Predict values for the label column of new data.

        Parameters
        ----------
        data
            The data to make predictions for. Should contain same column names as training data and
            follow same format (except for the `label` column).
        candidate_data
            The candidate data from which to search the query data's matches.
        as_pandas
            Whether to return the output as a pandas DataFrame(Series) (True) or numpy array (False).
        realtime
            Whether to do realtime inference, which is efficient for small data (default False).
            If provided None, we would infer it on based on the data modalities
            and sample number.

        Returns
        -------
        Array of predictions, one corresponding to each row in given dataset.
        """
        self.on_predict_start()
        ret_type = LOGITS
        if candidate_data:
            pred = self._match_queries_and_candidates(
                query_data=data,
                candidate_data=candidate_data,
                return_prob=False,
            )
        else:
            outputs = self.predict_per_run(
                data=data,
                realtime=realtime,
                requires_label=False,
            )
            logits = extract_from_output(outputs=outputs, ret_type=ret_type)

            if self._df_preprocessor:
                pred = self._df_preprocessor.transform_prediction(
                    y_pred=logits,
                )
            else:
                if isinstance(logits, (torch.Tensor, np.ndarray)) and logits.ndim == 2:
                    pred = logits.argmax(axis=1)
                else:
                    pred = logits

        if (as_pandas is None and isinstance(data, pd.DataFrame)) or as_pandas is True:
            pred = self._as_pandas(data=data, to_be_converted=pred)

        return pred

    def predict_proba(
        self,
        data: Union[pd.DataFrame, dict, list],
        candidate_data: Optional[Union[pd.DataFrame, dict, list]] = None,
        as_pandas: Optional[bool] = None,
        as_multiclass: Optional[bool] = True,
        realtime: Optional[bool] = False,
        **kwargs,
    ):
        """
        Predict probabilities class probabilities rather than class labels.
        This is only for the classification. Calling it for regression will throw an exception.

        Parameters
        ----------
        data
            The data to make predictions for. Should contain same column names as training data and
              follow same format (except for the `label` column).
        candidate_data
            The candidate data from which to search the query data's matches.
        as_pandas
            Whether to return the output as a pandas DataFrame(Series) (True) or numpy array (False).
        as_multiclass
            Whether to return the probability of all labels or
            just return the probability of the positive class for binary classification problems.
        realtime
            Whether to do realtime inference, which is efficient for small data (default False).
            If provided None, we would infer it on based on the data modalities
            and sample number.

        Returns
        -------
        Array of predicted class-probabilities, corresponding to each row in the given data.
        When as_multiclass is True, the output will always have shape (#samples, #classes).
        Otherwise, the output will have shape (#samples,)
        """
        self.on_predict_start()
        assert self._problem_type not in [
            REGRESSION,
        ], f"Problem {self._problem_type} has no probability output."

        if candidate_data:
            prob = self._match_queries_and_candidates(
                query_data=data,
                candidate_data=candidate_data,
                return_prob=True,
            )
        else:
            outputs = self.predict_per_run(
                data=data,
                realtime=realtime,
                requires_label=False,
            )
            logits = extract_from_output(outputs=outputs, ret_type=LOGITS)
            prob = logits_to_prob(logits)

        if not as_multiclass:
            if self._problem_type == BINARY:
                prob = prob[:, 1]

        if (as_pandas is None and isinstance(data, pd.DataFrame)) or as_pandas is True:
            prob = self._as_pandas(data=data, to_be_converted=prob)

        return prob

    def extract_embedding(
        self,
        data: Union[pd.DataFrame, dict, list],
        return_masks: Optional[bool] = False,
        as_tensor: Optional[bool] = False,
        as_pandas: Optional[bool] = False,
        realtime: Optional[bool] = False,
        **kwargs,
    ):
        """
        Extract features for each sample, i.e., one row in the provided dataframe `data`.

        Parameters
        ----------
        data
            The data to extract embeddings for. Should contain same column names as training dataset and
            follow same format (except for the `label` column).
        return_masks
            If true, returns a mask dictionary, whose keys are the same as those in the features dictionary.
            If a sample has empty input in feature column `image_0`, the sample will has mask 0 under key `image_0`.
        as_tensor
            Whether to return a Pytorch tensor.
        as_pandas
            Whether to return the output as a pandas DataFrame (True) or numpy array (False).
        realtime
            Whether to do realtime inference, which is efficient for small data (default False).
            If provided None, we would infer it on based on the data modalities
            and sample number.

        Returns
        -------
        Array of embeddings, corresponding to each row in the given data.
        It will have shape (#samples, D) where the embedding dimension D is determined
        by the neural network's architecture.
        """
        self.on_predict_start()
        turn_on_off_feature_column_info(
            data_processors=self._data_processors,
            flag=True,
        )
        outputs = self.predict_per_run(
            data=data,
            realtime=realtime,
            requires_label=False,
        )
        if self._problem_type in [FEATURE_EXTRACTION, ZERO_SHOT_IMAGE_CLASSIFICATION]:
            features = extract_from_output(outputs=outputs, ret_type=COLUMN_FEATURES, as_ndarray=as_tensor is False)
            if return_masks:
                masks = extract_from_output(outputs=outputs, ret_type=MASKS, as_ndarray=as_tensor is False)
        else:
            features = extract_from_output(outputs=outputs, ret_type=FEATURES, as_ndarray=as_tensor is False)

        if as_pandas:
            features = pd.DataFrame(features, index=data.index)
            if return_masks:
                masks = pd.DataFrame(masks, index=data.index)

        if return_masks:
            return features, masks
        else:
            return features

    def _as_pandas(
        self,
        data: Union[pd.DataFrame, dict, list],
        to_be_converted: Union[np.ndarray, dict],
    ):
        if isinstance(data, pd.DataFrame):
            index = data.index
        else:
            index = None
        if isinstance(to_be_converted, list) or (
            isinstance(to_be_converted, np.ndarray) and to_be_converted.ndim == 1
        ):
            return pd.Series(to_be_converted, index=index, name=self._label_column)
        else:
            return pd.DataFrame(to_be_converted, index=index, columns=self.class_labels)

    def _load_state_dict(
        self,
        state_dict: dict = None,
        path: str = None,
        prefix: str = "model.",
        strict: bool = True,
    ):
        if state_dict is None:
            if os.path.isdir(path + "-dir"):  # deepspeed save checkpoints into a directory
                from lightning.pytorch.utilities.deepspeed import convert_zero_checkpoint_to_fp32_state_dict

                convert_zero_checkpoint_to_fp32_state_dict(path + "-dir", path)
                shutil.rmtree(path + "-dir")
                state_dict = torch.load(path, map_location=torch.device("cpu"), weights_only=False)["state_dict"]  # nosec B614
            else:
                state_dict = torch.load(path, map_location=torch.device("cpu"), weights_only=False)["state_dict"]  # nosec B614
        state_dict = {k.partition(prefix)[2]: v for k, v in state_dict.items() if k.startswith(prefix)}

        # Some buffers like `position_ids` are registered as persistent=False since transformers 4.31.0
        # Refer to https://github.com/huggingface/transformers/pull/24505/files
        buffer_names = [k for k, v in self._model.named_buffers()]
        buffer_names_to_filter = [k for k in buffer_names if k not in self._model.state_dict().keys()]
        state_dict = {k: v for k, v in state_dict.items() if k not in buffer_names_to_filter}

        load_result = self._model.load_state_dict(state_dict, strict=strict)
        assert (
            len(load_result.unexpected_keys) == 0
        ), f"Load model failed, unexpected keys {load_result.unexpected_keys.__str__()}"

    @staticmethod
    def _replace_model_name_prefix(
        state_dict: dict,
        old_prefix: str,
        new_prefix: str,
    ):
        start_idx = len(old_prefix)
        state_dict_processed = {
            new_prefix + k[start_idx:]: v for k, v in state_dict.items() if k.startswith(old_prefix)
        }
        return state_dict_processed

    def save(
        self,
        path: str,
        standalone: Optional[bool] = True,
        config: Optional[DictConfig] = None,
        model: Optional[nn.Module] = None,
        df_preprocessor: Optional[MultiModalFeaturePreprocessor] = None,
        data_processors: Optional[Dict] = None,
        fit_called: Optional[bool] = None,
        save_model: Optional[bool] = True,
    ):
        """
        Save this learner to file in directory specified by `path`.

        Parameters
        ----------
        path
            The directory to save this learner.
        standalone
            Whether to save the downloaded model for offline deployment.
            When standalone = True, save the transformers.CLIPModel and transformers.AutoModel to os.path.join(path,model_name),
            and reset the associate model.model_name.checkpoint_name start with `local://` in config.yaml.
            When standalone = False, the saved artifact may require an online environment to process in load().
        """
        config = config if config else self._config
        config = copy.deepcopy(config)
        model = model if model else self._model
        if standalone and not config.optim.peft:
            config = save_pretrained_model_configs(model=model, config=config, path=path)
        os.makedirs(path, exist_ok=True)
        OmegaConf.save(config=config, f=os.path.join(path, "config.yaml"))

        df_preprocessor = df_preprocessor if df_preprocessor else self._df_preprocessor
        with open(os.path.join(path, "df_preprocessor.pkl"), "wb") as fp:
            pickle.dump(df_preprocessor, fp)

        data_processors = data_processors if data_processors else self._data_processors
        data_processors = copy.deepcopy(data_processors)

        # Save text tokenizers before saving data processors
        for modality in [TEXT, TEXT_NER, NER, DOCUMENT]:
            if modality in data_processors:
                for per_processor in data_processors[modality]:
                    per_processor.save_tokenizer(path)

        # Clear the documents cache dictionary before saving.
        for modality in [DOCUMENT]:
            if modality in data_processors:
                for p in data_processors[modality]:
                    p.documents.clear()

        with open(os.path.join(path, "data_processors.pkl"), "wb") as fp:
            pickle.dump(data_processors, fp)

        with open(os.path.join(path, "eval_metric.pkl"), "wb") as fp:
            pickle.dump(self._eval_metric_func, fp)

        with open(os.path.join(path, f"assets.json"), "w") as fp:
            json.dump(
                {
                    "learner_class": self.__class__.__name__,
                    "column_types": self._column_types,
                    "label_column": self._label_column,
                    "problem_type": self._problem_type,
                    "presets": self._presets,
                    "eval_metric_name": self._eval_metric_name,
                    "validation_metric_name": self._validation_metric_name,
                    "minmax_mode": self._minmax_mode,
                    "output_shape": self._output_shape,
                    "save_path": path,
                    "pretrained": self._pretrained,
                    "pretrained_path": self._pretrained_path,
                    "fit_called": fit_called if fit_called is not None else self._fit_called,
                    "best_score": self._best_score,
                    "total_train_time": self._total_train_time,
                    "version": ag_version.__version__,
                },
                fp,
                ensure_ascii=True,
            )

        if save_model:
            checkpoint = {"state_dict": {"model." + name: param for name, param in model.state_dict().items()}}
            torch.save(checkpoint, os.path.join(os.path.abspath(path), MODEL_CHECKPOINT))  # nosec B614

    @staticmethod
    def _load_metadata(
        learner: BaseLearner,
        path: str,
        resume: Optional[bool] = False,
        verbosity: Optional[int] = 3,
    ):
        path = os.path.abspath(os.path.expanduser(path))
        assert os.path.isdir(path), f"'{path}' must be an existing directory."
        config = OmegaConf.load(os.path.join(path, "config.yaml"))

        config = get_local_pretrained_config_paths(
            config=config, path=path
        )  # check the config to load offline pretrained model configs

        with open(os.path.join(path, "assets.json"), "r") as fp:
            assets = json.load(fp)

        with open(os.path.join(path, "df_preprocessor.pkl"), "rb") as fp:
            df_preprocessor = pickle.load(fp)  # nosec B301
        try:
            with open(os.path.join(path, "data_processors.pkl"), "rb") as fp:
                data_processors = pickle.load(fp)  # nosec B301
            # Load text tokenizers after loading data processors.
            for modality in [TEXT, TEXT_NER, NER, DOCUMENT]:
                if modality in data_processors:
                    for per_processor in data_processors[modality]:
                        per_processor.load_tokenizer(path)

            # Only keep the modalities with non-empty processors.
            data_processors = {k: v for k, v in data_processors.items() if len(v) > 0}
        except:  # reconstruct the data processor in case something went wrong.
            data_processors = None

        learner._label_column = assets["label_column"]
        learner._problem_type = assets["problem_type"]
        learner._presets = assets["presets"]
        learner._best_score = assets["best_score"]
        learner._total_train_time = assets["total_train_time"]
        learner._eval_metric_name = assets["eval_metric_name"]
        with open(os.path.join(path, "eval_metric.pkl"), "rb") as fp:
            learner._eval_metric_func = pickle.load(fp)  # nosec B301
        learner._verbosity = verbosity
        learner._resume = resume
        learner._save_path = path  # in case the original exp dir is copied to somewhere else
        learner._pretrained_path = path
        learner._pretrained = assets["pretrained"]
        learner._fit_called = assets["fit_called"]
        learner._config = config
        learner._output_shape = assets["output_shape"]
        if "classes" in assets:
            learner._classes = assets["classes"]
        learner._column_types = assets["column_types"]
        learner._validation_metric_name = assets["validation_metric_name"]
        learner._df_preprocessor = df_preprocessor
        learner._data_processors = data_processors
        learner._minmax_mode = assets["minmax_mode"]

        return learner

    @classmethod
    def load(
        cls,
        path: str,
        resume: Optional[bool] = False,
        verbosity: Optional[int] = 3,
    ):
        """
        Load a learner object from a directory specified by `path`. The to-be-loaded learner
        can be completely or partially trained by .fit(). If a previous training has completed,
        it will load the checkpoint `model.ckpt`. Otherwise if a previous training accidentally
        collapses in the middle, it can load the `last.ckpt` checkpoint by setting `resume=True`.
        It also supports loading one specific checkpoint given its path.

        Parameters
        ----------
        path
            The directory to load the learner object.
        resume
            Whether to resume training from `last.ckpt`. This is useful when a training was accidentally
            broken during the middle and we want to resume the training from the last saved checkpoint.
        verbosity
            Verbosity levels range from 0 to 4 and control how much information is printed.
            Higher levels correspond to more detailed print statements (you can set verbosity = 0 to suppress warnings).

        Returns
        -------
        The loaded learner object.
        """
        dir_path, ckpt_path = get_dir_ckpt_paths(path=path)

        assert os.path.isdir(dir_path), f"'{dir_path}' must be an existing directory."
        learner = cls(label="dummy_label")
        learner = cls._load_metadata(learner=learner, path=dir_path, resume=resume, verbosity=verbosity)
        peft = learner._config.optim.peft
        learner._model = create_fusion_model(
            config=learner._config,
            num_classes=learner._output_shape,
            classes=learner._classes if hasattr(learner, "_classes") else None,
            num_numerical_columns=len(learner._df_preprocessor.numerical_feature_names),
            num_categories=learner._df_preprocessor.categorical_num_categories,
            pretrained=False if not peft else True,  # set "pretrain=False" to prevent downloading online models
        )
        if learner._data_processors is None:
            learner._data_processors = create_fusion_data_processors(
                config=learner._config,
                model=learner._model,
            )
        load_path, ckpt_path = get_load_ckpt_paths(
            ckpt_path=ckpt_path,
            dir_path=dir_path,
            resume=resume,
        )
        learner._load_state_dict(
            path=load_path,
            strict=not peft,
        )
        learner._ckpt_path = ckpt_path
        loss_func = get_loss_func(
            problem_type=learner._problem_type,
            mixup_active=False,
            loss_func_name=learner._config.optim.loss_func,
            config=learner._config.optim,
            num_classes=learner._output_shape,
        )
        model_postprocess_fn = get_model_postprocess_fn(
            problem_type=learner._problem_type,
            loss_func=loss_func,
        )
        learner._model_postprocess_fn = model_postprocess_fn
        learner._config = learner.update_strategy_by_env(learner._config)

        return learner

    @property
    def class_labels(self):
        """
        The original name of the class labels.
        For example, the tabular data may contain classes equal to
        "entailment", "contradiction", "neutral". Internally, these will be converted to
        0, 1, 2, ...
        This function returns the original names of these raw labels.

        Returns
        -------
        List that contain the class names. It will be None if it's not a classification problem.
        """
        if self._problem_type == MULTICLASS or self._problem_type == BINARY:
            return self._df_preprocessor.label_generator.classes_
        else:
            warnings.warn("Accessing class names for a non-classification problem. Return None.")
            return None

    @property
    def positive_class(self):
        """
        Name of the class label that will be mapped to 1.
        This is only meaningful for binary classification problems.

        It is useful for computing metrics such as F1 which require a positive and negative class.
        You may refer to https://en.wikipedia.org/wiki/F-score for more details.
        In binary classification, :class:`BaseLearner.predict_proba(as_multiclass=False)`
        returns the estimated probability that each row belongs to the positive class.
        Will print a warning and return None if called when `learner.problem_type != 'binary'`.

        Returns
        -------
        The positive class name in binary classification or None if the problem is not binary classification.
        """
        if self._problem_type != BINARY:
            logger.warning(
                f"Warning: Attempted to retrieve positive class label in a non-binary problem. "
                f"Positive class labels only exist in binary classification. "
                f"Returning None instead. The problem type is '{self._problem_type}'"
                f" but positive_class only exists for '{BINARY}'."
            )
            return None
        else:
            return self.class_labels[1]

    def fit_summary(self, verbosity=0, show_plot=False):
        """
        Output summary of information about models produced during `fit()`.

        Parameters
        ----------
        verbosity : int, default = 2
            Verbosity levels range from 0 to 4 and control how much information is printed.
            verbosity = 0 for no output printing.
            TODO: Higher levels correspond to more detailed print statements
        show_plot : bool, default = False
            If True, shows the model summary plot in browser when verbosity > 1.

        Returns
        -------
        Dict containing various detailed information.
        We do not recommend directly printing this dict as it may be very large.
        """
        if self._total_train_time is None:
            logging.info("There is no `best_score` or `total_train_time`. Have you called `predictor.fit()`?")
        else:
            logging.info(
                f"Here's the model summary:"
                f""
                f"The model achieved score '{self._best_score}' on the validation metric"
                f" '{self._validation_metric_name}'. "
                f"The total training time is {timedelta(seconds=self._total_train_time)}"
            )
        results = {
            f"val_{self._validation_metric_name}": self._best_score,
            "training_time": self._total_train_time,
        }
        return results

    def list_supported_models(self, pretrained=True):
        """
        List supported models for each problem_type to let users know
        options of checkpoint name to choose during fit().

        Parameters
        ----------
        pretrained : bool, default = True
            If True, only return the models with pretrained weights.
            If False, return all the models as long as there is model definition.

        Returns
        -------
        a list of model names
        """
        if self.problem_property and self.problem_property.is_classification:
            # FIXME (Need to list the supported models for each modality)
            return list_timm_models(pretrained=pretrained)
        else:
            raise ValueError(f"list_supported_models() is not available for problem type: {self._problem_type}")

    @staticmethod
    def update_strategy_by_env(config):
        """
        Set strategy to ddp_fork or ddp_notebook if an iterative env is detected.
        """
        assert config is not None
        if is_interactive_env() and not is_interactive_strategy(config.env.strategy):
            strs = list(config.env.strategy.partition("_find_unused_parameters"))
            strs[0] = "ddp_fork"
            config.env.strategy = "".join(strs)

        return config

    def set_num_gpus(self, num_gpus):
        assert isinstance(num_gpus, int)
        self._config.env.num_gpus = num_gpus

    def get_num_gpus(self):
        try:
            return self._config.env.num_gpus
        except:
            return None



import copy
import logging
import os
import re
import warnings
from typing import Callable, Dict, List, Optional, Tuple, Union

from omegaconf import DictConfig, ListConfig, OmegaConf
from torch import nn

from ..constants import DATA, FT_TRANSFORMER, FUSION_TRANSFORMER, HF_MODELS, MODEL, REGRESSION, VALID_CONFIG_KEYS
from .presets import get_basic_config, get_ensemble_presets, get_presets

logger = logging.getLogger(__name__)


def filter_search_space(hyperparameters: Dict, keys_to_filter: Union[str, List[str]]):
    """
    Filter search space within hyperparameters without the given keys as prefixes.
    Hyperparameters that are not search space will not be filtered.

    Parameters
    ----------
    hyperparameters
        A dictionary containing search space and overrides to config.
    keys_to_filter
        Keys that needs to be filtered out

    Returns
    -------
        hyperparameters being filtered
    """
    if isinstance(keys_to_filter, str):
        keys_to_filter = [keys_to_filter]

    assert any(
        key.startswith(valid_keys) for valid_keys in VALID_CONFIG_KEYS for key in keys_to_filter
    ), f"Invalid keys: {keys_to_filter}. Valid options are {VALID_CONFIG_KEYS}"
    from ray.tune.search.sample import Domain

    from autogluon.common import space

    hyperparameters = copy.deepcopy(hyperparameters)
    for hyperparameter, value in hyperparameters.copy().items():
        if not isinstance(value, (space.Space, Domain)):
            continue
        for key in keys_to_filter:
            if hyperparameter.startswith(key):
                del hyperparameters[hyperparameter]
    return hyperparameters


def get_default_config(config: Optional[Union[Dict, DictConfig]] = None, extra: Optional[List[str]] = None):
    """
    Get the default config.

    Parameters
    ----------
    config
        A dictionary including four keys: "model", "data", "optim", and "env".
        If any key is not given, we will fill in with the default value.
    extra
        A list of extra config keys.

    Returns
    -------
    The default config.
    """
    if isinstance(config, DictConfig):
        return config

    if config is None:
        config = {}

    basic_config = get_basic_config(extra=extra)
    for k, default_value in basic_config.items():
        if k not in config:
            config[k] = default_value

    all_configs = []
    for k, v in config.items():
        if isinstance(v, dict):
            per_config = OmegaConf.create(v)
        elif isinstance(v, DictConfig):
            per_config = v
        elif isinstance(v, str):
            if v.lower().endswith((".yaml", ".yml")):
                per_config = OmegaConf.load(os.path.expanduser(v))
            else:
                cur_path = os.path.dirname(os.path.dirname(os.path.abspath(__file__)))
                config_path = os.path.join(cur_path, "configs", k, f"{v}.yaml")
                per_config = OmegaConf.load(config_path)
        else:
            raise ValueError(f"Unknown configuration type: {type(v)}")

        all_configs.append(per_config)

    config = OmegaConf.merge(*all_configs)

    return config


def get_config(
    problem_type: Optional[str] = None,
    presets: Optional[str] = None,
    config: Optional[Union[Dict, DictConfig]] = None,
    overrides: Optional[Union[str, List[str], Dict]] = None,
    extra: Optional[List[str]] = None,
):
    """
    Construct configurations for model, data, optim, and env.
    It supports to overrides some default configurations.

    Parameters
    ----------
    problem_type
        Problem type.
    presets
        Presets regarding model quality, e.g., best_quality, high_quality, and medium_quality.
    config
        A dictionary including four keys: "model", "data", "optim", and "env".
        If any key is not given, we will fill in with the default value.

        The value of each key can be a string, yaml path, or DictConfig object. For example:
        config = {
                        "model": "default",
                        "data": "default",
                        "optim": "default",
                        "env": "default",
                    }
            or
            config = {
                        "model": "/path/to/model/config.yaml",
                        "data": "/path/to/data/config.yaml",
                        "optim": "/path/to/optim/config.yaml",
                        "env": "/path/to/env/config.yaml",
                    }
            or
            config = {
                        "model": OmegaConf.load("/path/to/model/config.yaml"),
                        "data": OmegaConf.load("/path/to/data/config.yaml"),
                        "optim": OmegaConf.load("/path/to/optim/config.yaml"),
                        "env": OmegaConf.load("/path/to/env/config.yaml"),
                    }
    overrides
        This is to override some default configurations.
            For example, changing the text and image backbones can be done by formatting:

            a string
            overrides = "model.hf_text.checkpoint_name=google/electra-small-discriminator
            model.timm_image.checkpoint_name=swin_small_patch4_window7_224"

            or a list of strings
            overrides = ["model.hf_text.checkpoint_name=google/electra-small-discriminator",
            "model.timm_image.checkpoint_name=swin_small_patch4_window7_224"]

            or a dictionary
            overrides = {
                            "model.hf_text.checkpoint_name": "google/electra-small-discriminator",
                            "model.timm_image.checkpoint_name": "swin_small_patch4_window7_224",
                        }
    extra
        A list of extra config keys.

    Returns
    -------
    Configurations as a DictConfig object
    """

    if not config and not presets:
        presets = "default"

    if not isinstance(config, DictConfig):
        if presets is None:
            preset_overrides = None
        else:
            preset_overrides, _ = get_presets(problem_type=problem_type, presets=presets)

        config = get_default_config(config, extra=extra)
        # apply the preset's overrides
        if preset_overrides:
            config = apply_omegaconf_overrides(config, overrides=preset_overrides, check_key_exist=True)

    verify_model_names(config.model)
    logger.debug(f"overrides: {overrides}")
    if overrides is not None:
        # avoid manipulating the user-provided overrides
        overrides = copy.deepcopy(overrides)
        # apply customized model names
        overrides = parse_dotlist_conf(overrides)  # convert to a dict
        config.model, _ = customize_model_names(
            config=config.model,
            customized_names=overrides.get("model.names", None),
        )
        # remove `model.names` from overrides since it's already applied.
        overrides.pop("model.names", None)
        # apply the user-provided overrides
        config = apply_omegaconf_overrides(config, overrides=overrides, check_key_exist=True)
    verify_model_names(config.model)
    return config


def verify_model_names(config: DictConfig):
    """
    Verify whether provided model names are valid.

    Parameters
    ----------
    config
        Config should have a attribute `names`, which contains a list of
        attribute names, e.g., ["timm_image", "hf_text"]. And each string in
        `config.names` should also be a attribute of `config`, e.g, `config.timm_image`.
    """
    # must have attribute `names`
    assert hasattr(config, "names")
    # return if no names available
    if not config.names:
        return
    # assure no duplicate names
    assert len(config.names) == len(set(config.names))
    # verify that strings in `config.names` match the keys of `config`.
    keys = list(config.keys())
    keys.remove("names")
    assert set(config.names).issubset(set(keys)), f"`{config.names}` do not match config keys {keys}"

    # verify that no name starts with another one
    names = sorted(config.names, key=lambda ele: len(ele), reverse=True)
    for i in range(len(names)):
        if names[i].startswith(tuple(names[i + 1 :])):
            raise ValueError(f"name {names[i]} starts with one of another name: {names[i+1:]}")


def get_name_prefix(
    name: str,
    prefixes: List[str],
):
    """
    Get a name's prefix from some available candidates.

    Parameters
    ----------
    name
        A name string
    prefixes
        Available prefixes.

    Returns
    -------
        Prefix of the name.
    """
    search_results = [pre for pre in prefixes if name.lower().startswith(pre)]
    if len(search_results) == 0:
        return None
    elif len(search_results) >= 2:
        raise ValueError(
            f"Model name `{name}` is mapped to multiple models, "
            f"which means some names in `{prefixes}` have duplicate prefixes."
        )
    else:
        return search_results[0]


def customize_model_names(
    config: DictConfig,
    customized_names: Union[str, List[str]],
    advanced_hyperparameters: Optional[Dict] = None,
):
    """
    Customize attribute names of `config` with the provided names.
    A valid customized name string should start with one available name
    string in `config`. Customizing the model names in advanced_hyperparameters
    is only used for matcher query and response models currently.

    Parameters
    ----------
    config
        Config should have a attribute `names`, which contains a list of
        attribute names, e.g., ["timm_image", "hf_text"]. And each string in
        `config.names` should also be a attribute of `config`, e.g, `config.timm_image`.
    customized_names
        The provided names to replace the existing ones in `config.names` as well as
        the corresponding attribute names. For example, if `customized_names` is
        ["timm_image_123", "hf_text_abc"], then `config.timm_image` and `config.hf_text`
        are changed to `config.timm_image_123` and `config.hf_text_abc`.
    advanced_hyperparameters
        The hyperparameters whose values are complex objects, which can't be stored in config.

    Returns
    -------
        A new config with its first-level attributes customized by the provided names.
    """
    if not customized_names:
        return config, advanced_hyperparameters

    if isinstance(customized_names, str):
        customized_names = OmegaConf.from_dotlist([f"names={customized_names}"]).names

    if advanced_hyperparameters:
        new_advanced_hyperparameters = copy.deepcopy(advanced_hyperparameters)
    else:
        new_advanced_hyperparameters = dict()

    new_config = OmegaConf.create()
    new_config.names = []
    available_prefixes = list(config.keys())
    available_prefixes.remove("names")
    for per_name in customized_names:
        per_prefix = get_name_prefix(
            name=per_name,
            prefixes=available_prefixes,
        )
        if per_prefix:
            per_config = getattr(config, per_prefix)
            setattr(new_config, per_name, copy.deepcopy(per_config))
            new_config.names.append(per_name)

            if advanced_hyperparameters:
                for k, v in advanced_hyperparameters.items():
                    if k.startswith(f"{MODEL}.{per_prefix}"):
                        new_k = k.replace(f"{MODEL}.{per_prefix}", f"{MODEL}.{per_name}")
                        new_advanced_hyperparameters.pop(k)
                        new_advanced_hyperparameters[new_k] = v
        else:
            logger.debug(f"Removing {per_name}, which doesn't start with any of these prefixes: {available_prefixes}.")

    if len(new_config.names) == 0:
        raise ValueError(
            f"No customized name in `{customized_names}` starts with name prefixes in `{available_prefixes}`."
        )

    return new_config, new_advanced_hyperparameters


def save_pretrained_model_configs(
    model: nn.Module,
    config: DictConfig,
    path: str,
) -> DictConfig:
    """
    Save the pretrained model configs to local to make future loading not dependent on Internet access.
    By initializing models with local configs, Huggingface doesn't need to download pretrained weights from Internet.

    Parameters
    ----------
    model
        One model.
    config
        A DictConfig object. The model config should be accessible by "config.model".
    path
        The path to save pretrained model configs.
    """
    # TODO? Fix hardcoded model names.
    requires_saving = any([model_name.lower().startswith(HF_MODELS) for model_name in config.model.names])
    if not requires_saving:
        return config

    if (
        len(config.model.names) == 1
    ):  # TODO: Not sure this is a sufficient check. Hyperparameter "model.names" : ["hf_text", "fusion_mlp"] fails here.
        model = nn.ModuleList([model])
    else:  # assumes the fusion model has a model attribute, a nn.ModuleList
        model = model.model
    for per_model in model:
        if per_model.prefix.lower().startswith(HF_MODELS):
            per_model.config.save_pretrained(os.path.join(path, per_model.prefix))
            model_config = getattr(config.model, per_model.prefix)
            model_config.checkpoint_name = os.path.join("local://", per_model.prefix)

    return config


def get_local_pretrained_config_paths(config: DictConfig, path: str) -> DictConfig:
    """
    Get the local config paths of hugginface pretrained models. With a local config,
    Hugginface can initialize a model without having to download its pretrained weights.

    Parameters
    ----------
    config
        A DictConfig object. The model config should be accessible by "config.model".
    path
        The saving path to the pretrained model configs.
    """
    for model_name in config.model.names:
        if model_name.lower().startswith(HF_MODELS):
            model_config = getattr(config.model, model_name)
            if model_config.checkpoint_name.startswith("local://"):
                model_config.checkpoint_name = os.path.join(path, model_config.checkpoint_name[len("local://") :])
                assert os.path.exists(
                    os.path.join(model_config.checkpoint_name, "config.json")
                )  # guarantee the existence of local configs

    return config


def parse_dotlist_conf(conf):
    """
    Parse the config files that is potentially in the dotlist format to a dictionary.

    Parameters
    ----------
    conf
        Apply the conf stored as dotlist, e.g.,
         'aaa=a, bbb=b' or ['aaa=a, ', 'bbb=b'] to {'aaa': 'a', 'bbb': b}

    Returns
    -------
    new_conf
    """
    if isinstance(conf, str):
        conf = conf.split()
        need_parse = True
    elif isinstance(conf, (list, tuple)):
        need_parse = True
    elif isinstance(conf, dict):
        need_parse = False
    else:
        raise ValueError(f"Unsupported format of conf={conf}")
    if need_parse:
        new_conf = dict()
        curr_key = None
        curr_value = ""
        for ele in conf:
            if "=" in ele:
                key, v = ele.split("=")
                if curr_key is not None:
                    new_conf[curr_key] = curr_value
                curr_key = key
                curr_value = v
            else:
                if curr_key is None:
                    raise ValueError(f"Cannot parse the conf={conf}")
                curr_value = curr_value + " " + ele
        if curr_key is not None:
            new_conf[curr_key] = curr_value
        return new_conf
    else:
        return conf


def apply_omegaconf_overrides(
    conf: DictConfig,
    overrides: Union[List, Tuple, str, Dict, DictConfig],
    check_key_exist=True,
):
    """
    Apply omegaconf overrides.

    Parameters
    ----------
    conf
        The base configuration.
    overrides
        The overrides can be a string or a list.
    check_key_exist
        Whether to check if all keys in the overrides must exist in the conf.

    Returns
    -------
    new_conf
        The updated configuration.
    """
    overrides = parse_dotlist_conf(overrides)
    overrides = make_overrides_backward_compatible(overrides)

    def _check_exist_dotlist(C, key_in_dotlist):
        if not isinstance(key_in_dotlist, list):
            key_in_dotlist = key_in_dotlist.split(".")
        if key_in_dotlist[0] in C:
            if len(key_in_dotlist) > 1:
                return _check_exist_dotlist(C[key_in_dotlist[0]], key_in_dotlist[1:])
            else:
                return True
        else:
            return False

    if check_key_exist:
        for ele in overrides.items():
            if not _check_exist_dotlist(conf, ele[0]):
                raise KeyError(
                    f'"{ele[0]}" is not found in the config. You may need to check the overrides. '
                    f"overrides={overrides}"
                )
    override_conf = OmegaConf.from_dotlist([f"{ele[0]}={ele[1]}" for ele in overrides.items()])
    replace_none_str(override_conf)
    conf = OmegaConf.merge(conf, override_conf)
    return conf


def replace_none_str(config: Union[DictConfig, ListConfig, dict, list]):
    """
    In-place replace "None" and "none" strings in the config with None.

    Parameters
    ----------
    config
        A config of type DictConfig, ListConfig, dict, or list.
    """
    if isinstance(config, (dict, DictConfig)):
        for key, value in config.items():
            if isinstance(value, str) and value.lower() == "none":
                config[key] = None
            elif isinstance(value, (dict, list, DictConfig, ListConfig)):
                replace_none_str(value)
    elif isinstance(config, (list, ListConfig)):
        for i, value in enumerate(config):
            if isinstance(value, str) and value.lower() == "none":
                config[i] = None
            elif isinstance(value, (dict, list, DictConfig, ListConfig)):
                replace_none_str(value)


def update_config_by_rules(
    problem_type: str,
    config: DictConfig,
):
    """
    Modify configs based on the need of loss func.
    Now it support changing the preprocessing of numerical label into Minmaxscaler while using BCEloss.

    Parameters
    ----------
    problem_type
        The type of the problem of the project.
    config
        The config of the project. It is a Dictconfig object.

    Returns
    -------
    The modified config.
    """
    loss_func = config.optim.loss_func
    if loss_func is not None:
        if problem_type == REGRESSION and "bce" in loss_func.lower():
            # To use BCELoss for regression problems, need to first scale the labels.
            config.data.label.numerical_preprocessing = "minmaxscaler"

    return config


def update_tabular_config_by_resources(
    config: DictConfig,
    num_numerical_columns: Optional[int] = 0,
    num_categorical_columns: Optional[int] = 0,
    resource: Optional[int] = 16,
):
    """
    Modify configs based on the dataset statistics.
    Use Additive attention with large column count and tune batch size accordingly.
    Parameters
    ----------
    config
        The config of the project. It is a Dictconfig object.
    num_numerical_columns
        The number of numerical columns.
    num_categorical_columns
        The number of categorical columns.
    resource
        The maximum resource (memory in GB) a single GPU has.
    Returns
    -------
    The modified config.
    """
    columns_per_model = {
        FUSION_TRANSFORMER: num_categorical_columns + num_numerical_columns,
        FT_TRANSFORMER: num_categorical_columns + num_numerical_columns,
    }

    # Threshold is expected to be ~= batch_size * num_tokens, for additive attention.
    # The multiplier 2e4 is a heuristic found from AutoML Benchmark.
    # TODO: determine the threshold/batch_size on training data directly
    threshold = resource * 2e4
    per_gpu_batch_size = config.env.per_gpu_batch_size
    for model in columns_per_model:
        if model in config.model.names:
            model_ = getattr(config.model, model)
            if columns_per_model[model] > 300 and model_.additive_attention == "auto":
                model_.additive_attention = True
                model_.share_qv_weights = True if model_.share_qv_weights == "auto" else model_.share_qv_weights
                warnings.warn(
                    f"Dataset contains >300 features, using additive attention for efficiency",
                    UserWarning,
                )
                if columns_per_model[model] * per_gpu_batch_size > threshold:
                    per_gpu_batch_size = int(threshold / columns_per_model[model])

            model_.additive_attention = False if model_.additive_attention == "auto" else model_.additive_attention
            model_.share_qv_weights = False if model_.share_qv_weights == "auto" else model_.share_qv_weights

    per_gpu_batch_size = max(per_gpu_batch_size, 1)
    if per_gpu_batch_size < config.env.per_gpu_batch_size:
        config.env.per_gpu_batch_size = per_gpu_batch_size
        warnings.warn(
            f"Setting  per_gpu_batch_size to {per_gpu_batch_size} to fit into GPU memory",
            UserWarning,
        )

    return config


def get_pretrain_configs_dir(subfolder: Optional[str] = None):
    import autogluon.multimodal

    pretrain_config_dir = os.path.join(autogluon.multimodal.__path__[0], "configs", "pretrain")
    if subfolder:
        pretrain_config_dir = os.path.join(pretrain_config_dir, subfolder)
    return pretrain_config_dir


def filter_timm_pretrained_cfg(cfg, remove_source=False, remove_null=True):
    filtered_cfg = {}
    keep_null = {"pool_size", "first_conv", "classifier"}  # always keep these keys, even if none
    for k, v in cfg.items():
        if remove_source and k in {"url", "file", "hf_hub_id", "hf_hub_id", "hf_hub_filename", "source"}:
            continue
        if remove_null and v is None and k not in keep_null:
            continue
        filtered_cfg[k] = v
    return filtered_cfg


def update_hyperparameters(
    problem_type,
    presets,
    provided_hyperparameters,
    provided_hyperparameter_tune_kwargs,
):
    """
    Update preset hyperparameters hyperparameter_tune_kwargs by the provided.
    Currently, this is mainly used for HPO presets, which define some searchable hyperparameters.
    We need to combine these searchable hyperparameters with ones provided by users.

    Parameters
    ----------
    problem_type
        Problem type.
    presets
        A preset string regarding modality quality or hpo.
    provided_hyperparameters
        The hyperparameters provided by users.
    provided_hyperparameter_tune_kwargs
        The hyperparameter_tune_kwargs provided by users.

    Returns
    -------
    The updated hyperparameters and hyperparameter_tune_kwargs.
    """
    hyperparameters, hyperparameter_tune_kwargs = get_presets(problem_type=problem_type, presets=presets)

    if hyperparameter_tune_kwargs and provided_hyperparameter_tune_kwargs:
        hyperparameter_tune_kwargs.update(provided_hyperparameter_tune_kwargs)
    elif provided_hyperparameter_tune_kwargs:
        hyperparameter_tune_kwargs = provided_hyperparameter_tune_kwargs

    if hyperparameter_tune_kwargs:
        if provided_hyperparameters:
            hyperparameters.update(provided_hyperparameters)
    else:  # use the provided hyperparameters if no hpo. The preset hyperparameters will be also used later in get_config.
        hyperparameters = provided_hyperparameters

    if hyperparameter_tune_kwargs:
        assert isinstance(
            hyperparameters, dict
        ), "Please provide hyperparameters as a dictionary if you want to do HPO"

    return hyperparameters, hyperparameter_tune_kwargs


def filter_hyperparameters(
    hyperparameters: Dict,
    column_types: Dict,
    config: Optional[Union[Dict, DictConfig]] = None,
    fit_called: Optional[bool] = False,
):
    """
    Filter out the hyperparameters that have no effect for HPO.

    Parameters
    ----------
    hyperparameters
        The hyperparameters to override the default config.
    column_types
        Dataframe's column types.
    config
        A config provided by users or from the previous training.
    fit_called
        Whether fit() has been called.

    Returns
    -------
    The filtered hyperparameters.
    """
    model_names_key = f"{MODEL}.names"
    keys_to_filter = []

    if model_names_key in hyperparameters:
        # If continuous training or config is provided, make sure models are in config.model.
        config = get_default_config(config)
        selected_model_names = []
        config_model_names = list(config.model.keys())
        config_model_names.remove("names")
        for name in hyperparameters[model_names_key]:
            if name in config_model_names:
                selected_model_names.append(name)
        hyperparameters[model_names_key] = selected_model_names
        assert (
            len(selected_model_names) > 0
        ), f"hyperparameters['model.names'] {hyperparameters[model_names_key]} doesn't match any config model names {config_model_names}."

        # Filter models that are not in hyperparameters[model_names_key]
        # Avoid key not in config error when applying the overrides later.
        model_keys = [k for k in hyperparameters.keys() if k.startswith(MODEL)]
        if model_keys and model_names_key in model_keys:
            model_keys.remove(model_names_key)
            for k in model_keys:
                if k.split(".")[1] not in hyperparameters[model_names_key] and k not in keys_to_filter:
                    keys_to_filter.append(k)

        # Filter models whose data types are not detected.
        # Avoid sampling unused checkpoints, e.g., hf_text models for image classification, to run jobs,
        # which wastes resources and time.
        from ..data.utils import get_detected_data_types

        detected_data_types = get_detected_data_types(column_types)
        selected_model_names = []
        for model_name in hyperparameters[model_names_key]:
            model_config = config.model[model_name]
            if model_config.data_types:
                model_data_status = [d_type in detected_data_types for d_type in model_config.data_types]
                if not all(model_data_status):
                    keys_to_filter.append(f"{MODEL}.{model_name}")
                else:
                    selected_model_names.append(model_name)
            else:  # keep the fusion model, which will be handled by select_model().
                selected_model_names.append(model_name)
        hyperparameters[model_names_key] = selected_model_names
        assert (
            len(selected_model_names) > 0
        ), f"Model {hyperparameters[model_names_key]} can't handle the data with column types {column_types}"

    # Filter keys for continuous training.
    # Model and data processors would be reused.
    if fit_called:
        warnings.warn(
            "HPO while continuous training."
            "Hyperparameters related to Model and Data will NOT take effect."
            "We will filter them out from the search space."
        )
        keys_to_filter.extend([MODEL, DATA])

    for key in keys_to_filter:
        hyperparameters = {k: v for k, v in hyperparameters.items() if not k.startswith(key)}

    return hyperparameters


def split_hyperparameters(hyperparameters: Dict):
    """
    Split out some advanced hyperparameters whose values are complex objects instead of strings or numbers.

    Parameters
    ----------
    hyperparameters
        The user provided hyperparameters.

    Returns
    -------
    Hyperparameters and advanced hyperparameters.
    """
    if not isinstance(hyperparameters, dict):  # only support complex objects in dict.
        return hyperparameters, dict()

    if not hyperparameters:
        return hyperparameters, dict()

    advanced_hyperparameters = dict()
    for k, v in hyperparameters.items():
        if re.search("^model.*train_transforms$", k) or re.search("^model.*val_transforms$", k):
            if all([isinstance(trans, str) for trans in hyperparameters[k]]):
                pass
            elif all([isinstance(trans, Callable) for trans in hyperparameters[k]]):
                advanced_hyperparameters[k] = copy.deepcopy(v)
                hyperparameters[k] = str(v)  # get the objects' class strings
            else:
                raise ValueError(f"transform_types {v} contain neither all strings nor all callable objects.")

    return hyperparameters, advanced_hyperparameters


def update_ensemble_hyperparameters(
    presets,
    provided_hyperparameters,
):
    presets_hyperparameters, _ = get_ensemble_presets(presets=presets)
    if provided_hyperparameters:
        learner_names = provided_hyperparameters.pop("learner_names", None)
        if learner_names:
            assert isinstance(
                learner_names, list
            ), f"learner_names should be a list, but got type {type(learner_names)}"
            presets_hyperparameters = {k: v for k, v in presets_hyperparameters.items() if k in learner_names}
            provided_hyperparameters = {k: v for k, v in provided_hyperparameters.items() if k in learner_names}

        hyperparameters = copy.deepcopy(provided_hyperparameters)
        for k, v in presets_hyperparameters.items():
            if k not in hyperparameters:
                hyperparameters[k] = v
            else:
                for kk, vv in presets_hyperparameters[k].items():
                    if kk not in hyperparameters[k]:  # don't use presets to overwrite user-provided
                        hyperparameters[k][kk] = vv
    else:
        hyperparameters = presets_hyperparameters

    return hyperparameters


def make_overrides_backward_compatible(overrides: Dict):
    """
    Some config keys were changed in PR https://github.com/autogluon/autogluon/pull/4737
    This function is to make the changes backward compatible.

    Parameters
    ----------
    overrides
        A dictionary containing the user-provided hyperparameters,
        which may contain old config keys.

    Returns
    -------
    Overrides with up-to-date config keys.
    """
    key_pairs = {
        "optim.learning_rate": "optim.lr",
        "optim.efficient_finetune": "optim.peft",
        "optim.loss_function": "optim.loss_func",
        "env.num_workers_evaluation": "env.num_workers_inference",
        "env.eval_batch_size_ratio": "env.inference_batch_size_ratio",
        "data.label.numerical_label_preprocessing": "data.label.numerical_preprocessing",
        "model.categorical_mlp.drop_rate": "model.categorical_mlp.dropout",
        "model.numerical_mlp.drop_rate": "model.numerical_mlp.dropout",
        "model.numerical_mlp.d_token": "model.numerical_mlp.token_dim",
        "model.timm_image.max_img_num_per_col": "model.timm_image.max_image_num_per_column",
        "model.clip.max_img_num_per_col": "model.clip.max_image_num_per_column",
        "model.clip_image.max_img_num_per_col": "model.clip_image.max_image_num_per_column",
        "model.fusion_mlp.weight": "model.fusion_mlp.aux_loss_weight",
        "model.fusion_mlp.drop_rate": "model.fusion_mlp.dropout",
        "model.fusion_transformer.n_blocks": "model.fusion_transformer.num_blocks",
        "model.fusion_transformer.attention_n_heads": "model.fusion_transformer.attention_num_heads",
        "model.fusion_transformer.ffn_d_hidden": "model.fusion_transformer.ffn_hidden_size",
        "model.ft_transformer.attention_n_heads": "model.ft_transformer.attention_num_heads",
    }
    for k in list(overrides.keys()):
        provided_k = k
        if k.startswith("optimization."):
            k = "optim." + k[len("optimization.") :]
            logger.warning(
                f"The provided hyperparameter name {provided_k} contains a deprecated key `optimization.`. "
                f"Please replace `optimization.` with `optim.` when customizing the optimization hyperparameters."
            )

        if k in key_pairs:
            overrides[key_pairs[k]] = overrides.pop(provided_k)
            logger.warning(
                f"The hyperparameter name {provided_k} is depreciated. "
                f"We recommend using the new name {key_pairs[k]} instead."
                f"The deprecated hyperparameter will raise an exception starting in AutoGluon 1.4.0"
            )

    return overrides



import json
import logging
import os
from typing import Dict, List, Optional

import torch
from timm import create_model
from timm.layers.linear import Linear
from torch import nn

from ..constants import COLUMN, COLUMN_FEATURES, FEATURES, IMAGE, IMAGE_VALID_NUM, LABEL, LOGITS, MASKS
from .utils import (
    assign_layer_ids,
    get_column_features,
    get_image_size_mean_std,
    get_model_head,
    replace_missing_images_with_learnable,
)

logger = logging.getLogger(__name__)


# Stores the class names of the timm backbones that support variable input size. You can add more backbones to the list.
SUPPORT_VARIABLE_INPUT_SIZE_TIMM_CLASSES = {"convnext", "efficientnet", "mobilenetv3", "regnet", "resnet"}


class TimmAutoModelForImagePrediction(nn.Module):
    """
    Support TIMM image backbones.
    Refer to https://github.com/rwightman/pytorch-image-models
    """

    def __init__(
        self,
        prefix: str,
        checkpoint_name: str,
        num_classes: Optional[int] = 0,
        mix_choice: Optional[str] = "all_logits",
        pretrained: Optional[bool] = True,
        image_size: Optional[int] = None,
        image_norm: Optional[str] = None,
        image_chan_num: Optional[int] = 3,
        use_learnable_image: Optional[bool] = False,
    ):
        """
        Load a pretrained image backbone from TIMM.

        Parameters
        ----------
        prefix
            The prefix of the TimmAutoModelForImagePrediction model.
        checkpoint_name
            Name of the timm checkpoint, or local parent directory of the saved finetuned timm weights and config.
        num_classes
            The number of classes. 1 for a regression task.
        mix_choice
            Choice used for mixing multiple images. We now support.
            - all_images
                The images are directly averaged and passed to the model.
            - all_logits
                The logits output from individual images are averaged to generate the final output.
        pretrained
            Whether using the pretrained timm models. If pretrained=True, download the pretrained model.
        image_norm
            How to normalize an image. We now support:
            - inception
                Normalize image by IMAGENET_INCEPTION_MEAN and IMAGENET_INCEPTION_STD from timm
            - imagenet
                Normalize image by IMAGENET_DEFAULT_MEAN and IMAGENET_DEFAULT_STD from timm
            - clip
                Normalize image by mean (0.48145466, 0.4578275, 0.40821073) and
                std (0.26862954, 0.26130258, 0.27577711), used for CLIP.
        image_size
            The provided width / height of a square image.
        """
        super().__init__()
        # In TIMM, if num_classes==0, then create_model would automatically set self.model.head = nn.Identity()
        logger.debug(f"initializing {prefix} (TimmAutoModelForImagePrediction)")
        logger.debug(f"model checkpoint: {checkpoint_name}")
        if os.path.exists(checkpoint_name):
            checkpoint_path = f"{checkpoint_name}/pytorch_model.bin"
            try:
                with open(f"{checkpoint_name}/config.json") as f:
                    self.config = json.load(f)
                    pretrained_cfg = self.config.get("pretrained_cfg", {})
                    for k, v in pretrained_cfg.items():
                        if k not in self.config:
                            self.config[k] = v
                    self.checkpoint_name = self.config.get("architecture", None)
                    self.model = create_model(self.checkpoint_name, checkpoint_path=checkpoint_path, num_classes=0)
                    # create a head with new num_classes
                    self.head = (
                        Linear(in_features=self.config["num_features"], out_features=num_classes)
                        if num_classes > 0
                        else nn.Identity()
                    )
                    self.num_classes = num_classes if num_classes is not None else 0
            except:
                raise ValueError(f"Timm model path {checkpoint_name} does not exist or model is invalid.")
        else:
            self.checkpoint_name = checkpoint_name
            self.model = create_model(checkpoint_name, pretrained=pretrained, num_classes=num_classes)
            self.head = get_model_head(model=self.model)
            self.config = self.model.default_cfg
            self.num_classes = self.model.num_classes

        self.pretrained = pretrained
        self.out_features = self.model.num_features
        self.global_pool = self.model.global_pool if hasattr(self.model, "global_pool") else None
        self.model.reset_classifier(0)  # remove the internal head

        self.mix_choice = mix_choice
        logger.debug(f"mix_choice: {mix_choice}")

        self.prefix = prefix
        self.image_size, self.image_mean, self.image_std = get_image_size_mean_std(
            model_name=self.prefix,
            config=self.config,
            provided_size=image_size,
            provided_norm_type=image_norm,
            support_variable_input_size=self.support_variable_input_size(),
        )
        self.image_chan_num = image_chan_num
        self.use_learnable_image = use_learnable_image
        if self.use_learnable_image:
            self.learnable_image = nn.Parameter(torch.zeros(image_chan_num, self.image_size, self.image_size))
            logger.debug("will use a learnable image to replace missing ones")

        self.name_to_id = self.get_layer_ids()
        self.head_layer_names = [n for n, layer_id in self.name_to_id.items() if layer_id == 0]

    @property
    def image_key(self):
        return f"{self.prefix}_{IMAGE}"

    @property
    def image_valid_num_key(self):
        return f"{self.prefix}_{IMAGE_VALID_NUM}"

    @property
    def label_key(self):
        return f"{self.prefix}_{LABEL}"

    @property
    def input_keys(self):
        return [self.image_key, self.image_valid_num_key]

    @property
    def image_column_prefix(self):
        return f"{self.image_key}_{COLUMN}"

    @property
    def image_feature_dim(self):
        return self.model.num_features

    def support_variable_input_size(self):
        """Whether the TIMM image support images sizes that are different from the default used in the backbones"""
        if "test_input_size" in self.config and self.config["test_input_size"] != self.config["input_size"]:
            return True
        cls_name = type(self.model).__name__.lower()
        for k in SUPPORT_VARIABLE_INPUT_SIZE_TIMM_CLASSES:
            if cls_name in k:
                return True
        return False

    def forward(
        self,
        images: torch.FloatTensor,
        image_valid_num: torch.Tensor,
        image_column_names: Optional[List[str]] = None,
        image_column_indices: Optional[List[torch.Tensor]] = None,
    ):
        """
        Parameters
        ----------
        images : torch.FloatTensor
            A tensor in [N, C, H, W] layout to represent the images.
        image_valid_num : torch.Tensor
            A tensor that describes valid number of input images.
        image_column_names : list of str, optional
            A list of strings that indicates names of the image columns.
        image_column_indices : list of torch.Tensor, optional
            A list of tensors that indicates start and stop indices of the image columns.

        Returns
        -------
            A dictionary with logits and features.
        """
        column_features = column_feature_masks = dict()
        if self.mix_choice == "all_images":  # mix inputs
            mixed_images = (
                images.sum(dim=1) / torch.clamp(image_valid_num, min=1e-6)[:, None, None, None]
            )  # mixed shape: (b, 3, h, w)
            features = self.model(mixed_images)
            if self.num_classes > 0:
                logits = self.head(features)
            else:
                logits = features

        elif self.mix_choice == "all_logits":  # mix outputs
            b, n, c, h, w = images.shape
            steps = torch.arange(0, n).type_as(image_valid_num)
            image_masks = steps.reshape((1, -1)) < image_valid_num.reshape((-1, 1))  # (b, n)

            if self.use_learnable_image:
                images = replace_missing_images_with_learnable(
                    images=images,
                    image_masks=image_masks,
                    learnable_image=self.learnable_image,
                )
            features = self.model(images.reshape((b * n, c, h, w)))  # (b*n, num_features)
            if self.num_classes > 0:
                logits = self.head(features)
                logits = logits.reshape((b, n, -1))  # (b, n, num_classes)
            # reshape features after head prediction
            features = features.reshape((b, n, -1))  # (b, n, num_features)

            if not self.use_learnable_image:
                features = features * image_masks[:, :, None].type_as(features)  # (b, n, num_features)

            # need to collect column features before summing them
            if image_column_names:
                assert len(image_column_names) == len(image_column_indices), "invalid image column inputs"
                # collect features by image column names
                column_features, column_feature_masks = get_column_features(
                    batch=dict(zip(image_column_names, image_column_indices)),
                    column_name_prefix=self.image_column_prefix,
                    features=features,
                    valid_lengths=image_valid_num,
                )

            if self.use_learnable_image:
                features = features.mean(dim=1)
            else:
                features = features.sum(dim=1) / torch.clamp(image_valid_num, min=1e-6)[:, None]  # (b, num_features)
            if self.num_classes > 0:
                if self.use_learnable_image:
                    logits = logits.mean(dim=1)
                else:
                    logits = logits * image_masks[:, :, None].type_as(logits)  # (b, n, num_classes)
                    logits = logits.sum(dim=1) / torch.clamp(image_valid_num, min=1e-6)[:, None]  # (b, num_classes)
            else:
                logits = features

        else:
            raise ValueError(f"unknown mix_choice: {self.mix_choice}")

        return features, logits, column_features, column_feature_masks

    def get_output_dict(
        self,
        features: torch.Tensor,
        logits: torch.Tensor,
        column_features: Optional[Dict[str, torch.Tensor]] = None,
        column_feature_masks: Optional[Dict[str, torch.Tensor]] = None,
    ):
        ret = {COLUMN_FEATURES: {FEATURES: {}, MASKS: {}}}

        if column_features is not None and len(column_features) > 0:
            assert column_feature_masks is not None and len(column_features) == len(column_feature_masks)
            ret[COLUMN_FEATURES][FEATURES].update(column_features)
            ret[COLUMN_FEATURES][MASKS].update(column_feature_masks)

        ret[FEATURES] = features
        if self.num_classes > 0:
            ret[LOGITS] = logits

        return {self.prefix: ret}

    def get_layer_ids(
        self,
    ):
        """
        Assign an id to each layer. Layer ids will be used in layer-wise lr decay.
        Basically, id gradually increases when going from the output end to
        the input end. The layers defined in this class, e.g., head, have id 0.

        Due to different backbone architectures in TIMM, this function may not always return the correct result.
        Thus, you can use "print(json.dumps(name_to_id, indent=2))" to manually check whether
        the layer ids are reasonable.

        Returns
        -------
        A dictionary mapping the layer names (keys) to their ids (values).
        """
        model_prefix = "model"
        pre_encoder_patterns = ("embed", "cls_token", "stem", "bn1", "conv1")
        post_encoder_patterns = ("head", "norm", "bn2")
        names = [n for n, _ in self.named_parameters()]

        name_to_id, names = assign_layer_ids(
            names=names,
            pre_encoder_patterns=pre_encoder_patterns,
            post_encoder_patterns=post_encoder_patterns,
            model_pre=model_prefix,
        )

        if len(names) > 0:
            logger.debug(f"outer layers are treated as head: {names}")
        for n in names:
            assert n not in name_to_id
            name_to_id[n] = 0

        return name_to_id

    def dump_config(
        self,
        config_path: str,
    ):
        """
        Save TIMM image model configs to a local file.

        Parameters
        ----------
        config_path:
            A file to where the config is written to.
        """
        from ..utils import filter_timm_pretrained_cfg

        config = {}
        pretrained_cfg = filter_timm_pretrained_cfg(self.config, remove_source=True, remove_null=True)
        # set some values at root config level
        config["architecture"] = pretrained_cfg.pop("architecture")
        config["num_classes"] = self.num_classes
        config["num_features"] = self.out_features

        global_pool_type = getattr(self, "global_pool", None)
        if isinstance(global_pool_type, str) and global_pool_type:
            config["global_pool"] = global_pool_type

        config["pretrained_cfg"] = pretrained_cfg

        with open(config_path, "w") as f:
            json.dump(config, f, indent=2)
            logger.info(f"Timm config saved to {config_path}.")

    def save(self, save_path: str = "./", tokenizers: Optional[dict] = None):
        weights_path = f"{save_path}/pytorch_model.bin"
        torch.save(self.model.state_dict(), weights_path)  # nosec B614
        logger.info(f"Model {self.prefix} weights saved to {weights_path}.")
        config_path = f"{save_path}/config.json"
        self.dump_config(config_path)



import contextlib
import logging
import warnings
from typing import Dict, List, Optional, Tuple, Union

import torch

logger = logging.getLogger(__name__)


def convert_to_torch_precision(precision: Union[int, str]):
    """
    Convert a precision integer or string to the corresponding torch precision.

    Parameters
    ----------
    precision
    a precision integer or string from the config.

    Returns
    -------
    A torch precision object.
    """
    precision_mapping = {
        16: torch.half,
        "16": torch.half,
        "16-mixed": torch.half,
        "16-true": torch.half,
        "bf16": torch.bfloat16,
        "bf16-mixed": torch.bfloat16,
        "bf16-true": torch.bfloat16,
        32: torch.float32,
        "32": torch.float32,
        "32-true": torch.float32,
        64: torch.float64,
        "64": torch.float64,
        "64-true": torch.float64,
    }

    if precision in precision_mapping:
        precision = precision_mapping[precision]
    else:
        raise ValueError(f"Unknown precision: {precision}")

    return precision


def infer_precision(
    num_gpus: int, precision: Union[int, str], as_torch: Optional[bool] = False, cpu_only_warning: bool = True
):
    """
    Infer the proper precision based on the environment setup and the provided precision.

    Parameters
    ----------
    num_gpus
        GPU number.
    precision
        The precision provided in config.
    as_torch
        Whether to convert the precision to the Pytorch format.
    cpu_only_warning
        Whether to turn on warning if the instance has only CPU.

    Returns
    -------
    The inferred precision.
    """
    if num_gpus == 0:  # CPU only prediction
        if cpu_only_warning:
            warnings.warn(
                "Only CPU is detected in the instance. "
                "This may result in slow speed for MultiModalPredictor. "
                "Consider using an instance with GPU support.",
                UserWarning,
            )
        precision = 32  # Force to use fp32 for training since 16-mixed is not available in CPU
    else:
        if isinstance(precision, str) and "bf16" in precision and not torch.cuda.is_bf16_supported():
            warnings.warn(
                f"{precision} is not supported by the GPU device / cuda version. "
                "Consider using GPU devices with versions after Amphere or upgrading cuda to be >=11.0. "
                f"MultiModalPredictor is switching precision from {precision} to 32.",
                UserWarning,
            )
            precision = 32

    if as_torch:
        precision = convert_to_torch_precision(precision=precision)

    return precision


@contextlib.contextmanager
def double_precision_context():
    """
    Double precision context manager.
    """
    default_dtype = torch.get_default_dtype()
    torch.set_default_dtype(torch.float64)
    yield
    torch.set_default_dtype(default_dtype)


def get_precision_context(precision: Union[int, str], device_type: Optional[str] = None):
    """
    Choose the proper context manager based on the precision.

    Parameters
    ----------
    precision
        The precision.
    device_type
        gpu or cpu.

    Returns
    -------
    A precision context manager.
    """
    precision = convert_to_torch_precision(precision=precision)

    if precision in [torch.half, torch.float16, torch.bfloat16]:
        return torch.autocast(device_type=device_type, dtype=precision)
    if precision == torch.float32:
        assert torch.get_default_dtype() == torch.float32
        return contextlib.nullcontext()
    elif precision == torch.float64:
        return double_precision_context()
    else:
        raise ValueError(f"Unknown precision: {precision}")



import logging
from typing import Dict, List, Optional, Tuple

import torch
import torch.nn.functional as F
from torch import nn
from transformers import SamConfig

from ..constants import CLASS_LABEL, CLASS_LOGITS, COLUMN, IMAGE, IMAGE_VALID_NUM, LABEL, LOGITS, MASK_LABEL, MOE_LOSS
from .adaptation_layers import ConvLoRALinear
from .custom_hf_models.modeling_sam_for_conv_lora import SamImageSegmentationOutput, SamModel
from .utils import assign_layer_ids, freeze_model_layers, image_mean_std

logger = logging.getLogger(__name__)


def multi_class_mask_decoder_forward(
    self,
    image_embeddings: torch.Tensor,
    image_positional_embeddings: torch.Tensor,
    sparse_prompt_embeddings: torch.Tensor,
    dense_prompt_embeddings: torch.Tensor,
    multimask_output: bool,
    output_attentions: Optional[bool] = None,
    attention_similarity: torch.Tensor = None,
    target_embedding: torch.Tensor = None,
) -> Tuple[torch.Tensor, torch.Tensor]:
    """
    Modify the forward method of SamMaskDecoder for multi-class semantic segmentation
    based on https://github.com/huggingface/transformers/blob/main/src/transformers/models/sam/modeling_sam.py#L468.

    Args:
        image_embeddings (`torch.Tensor`):
            the embeddings from the image encoder
        image_positional_embedding (`torch.Tensor`):
            positional encoding with the shape of image_embeddings
        sparse_prompt_embeddings (`torch.Tensor`):
            The embeddings of the points and boxes
        dense_prompt_embeddings (`torch.Tensor`):
            the embeddings of the mask inputs
        multimask_output (bool):
            Whether to return multiple masks or a single mask.
        output_attentions (bool, *optional*):
            Whether or not to return the attentions tensors of all attention layers.
    """
    batch_size, num_channels, height, width = image_embeddings.shape
    point_batch_size = sparse_prompt_embeddings.shape[1]
    # Concatenate output tokens
    output_tokens = torch.cat([self.iou_token.weight, self.mask_tokens.weight], dim=0)
    output_tokens = output_tokens.repeat(batch_size, point_batch_size, 1, 1)

    if sparse_prompt_embeddings.sum().item() != 0:
        tokens = torch.cat((output_tokens, sparse_prompt_embeddings), dim=2)
    else:
        tokens = output_tokens
    point_embeddings = tokens.to(self.iou_token.weight.dtype)

    # Expand per-image data in batch direction to be per-point
    image_embeddings = image_embeddings + dense_prompt_embeddings
    image_embeddings = image_embeddings.repeat(point_batch_size, 1, 1, 1)
    image_positional_embeddings = image_positional_embeddings.repeat(point_batch_size, 1, 1, 1)

    # Run the transformer, image_positional_embedding are consumed
    point_embedding, image_embeddings, attentions = self.transformer(
        point_embeddings=point_embeddings,
        image_embeddings=image_embeddings,
        image_positional_embeddings=image_positional_embeddings,
        attention_similarity=attention_similarity,
        target_embedding=target_embedding,
        output_attentions=output_attentions,
    )
    iou_token_out = point_embedding[:, :, 0, :]
    mask_tokens_out = point_embedding[:, :, 1 : (1 + self.num_mask_tokens), :]

    # Upscale mask embeddings and predict masks using the mask tokens
    image_embeddings = image_embeddings.transpose(2, 3).reshape(
        batch_size * point_batch_size, num_channels, height, width
    )

    upscaled_embedding = self.upscale_conv1(image_embeddings)
    upscaled_embedding = self.activation(self.upscale_layer_norm(upscaled_embedding))
    upscaled_embedding = self.activation(self.upscale_conv2(upscaled_embedding))

    ################ Modify the original code. We aim at returning a single mask for each object.
    ################ Original logic is to return multiple masks.
    ################ So we use an MLP network to process all the mask proposals instead of multiple networks.
    # hyper_in_list = []
    # for i in range(self.num_mask_tokens):
    #     current_mlp = self.output_hypernetworks_mlps[i]
    #     hyper_in_list += [current_mlp(mask_tokens_out[:, :, i, :])]
    # hyper_in = torch.stack(hyper_in_list, dim=2)
    hyper_in = self.output_hypernetworks_mlps[0](mask_tokens_out)
    ################

    _, num_channels, height, width = upscaled_embedding.shape
    upscaled_embedding = upscaled_embedding.reshape(batch_size, point_batch_size, num_channels, height * width)
    masks = (hyper_in @ upscaled_embedding).reshape(
        batch_size, point_batch_size, -1, height, width
    )  # bs, 1, num_tokens, h, w

    ################ New added class prediction logic.
    class_predictions = self.output_classifier_mlps(mask_tokens_out).reshape(
        batch_size, point_batch_size, -1, self.num_classes + 1
    )
    ################

    # Generate mask quality predictions
    iou_pred = self.iou_prediction_head(iou_token_out)

    outputs = (masks, iou_pred)

    if output_attentions:
        outputs = outputs + (attentions,)
    else:
        outputs = outputs + (None,)

    return outputs + (class_predictions,)


def multi_class_sam_model_forward(
    self,
    pixel_values: Optional[torch.FloatTensor] = None,
    input_points: Optional[torch.FloatTensor] = None,
    input_labels: Optional[torch.LongTensor] = None,
    input_boxes: Optional[torch.FloatTensor] = None,
    input_masks: Optional[torch.LongTensor] = None,
    image_embeddings: Optional[torch.FloatTensor] = None,
    multimask_output: bool = True,
    attention_similarity: Optional[torch.FloatTensor] = None,
    target_embedding: Optional[torch.FloatTensor] = None,
    output_attentions: Optional[bool] = None,
    output_hidden_states: Optional[bool] = None,
    output_moe_loss: Optional[bool] = None,  # MoE loss for Conv-LoRA
    return_dict=None,
    **kwargs,
) -> List[Dict[str, torch.Tensor]]:
    r"""
    Modify the forward method of SamModel for multi-class semantic segmentation
    based on https://github.com/huggingface/transformers/blob/main/src/transformers/models/sam/modeling_sam.py#L1279.

    """
    output_attentions = output_attentions if output_attentions is not None else self.config.output_attentions
    output_hidden_states = (
        output_hidden_states if output_hidden_states is not None else self.config.output_hidden_states
    )
    return_dict = return_dict if return_dict is not None else self.config.use_return_dict

    if pixel_values is None and image_embeddings is None:
        raise ValueError("Either pixel_values or image_embeddings must be provided.")

    if pixel_values is not None and image_embeddings is not None:
        raise ValueError("Only one of pixel_values and image_embeddings can be provided.")

    if input_points is not None and len(input_points.shape) != 4:
        raise ValueError(
            "The input_points must be a 4D tensor. Of shape `batch_size`, `point_batch_size`, `nb_points_per_image`, `2`.",
            " got {}.".format(input_points.shape),
        )
    if input_boxes is not None and len(input_boxes.shape) != 3:
        raise ValueError(
            "The input_points must be a 3D tensor. Of shape `batch_size`, `nb_boxes`, `4`.",
            " got {}.".format(input_boxes.shape),
        )
    if input_points is not None and input_boxes is not None:
        point_batch_size = input_points.shape[1]
        box_batch_size = input_boxes.shape[1]
        if point_batch_size != box_batch_size:
            raise ValueError(
                "You should provide as many bounding boxes as input points per box. Got {} and {}.".format(
                    point_batch_size, box_batch_size
                )
            )

    image_positional_embeddings = self.get_image_wide_positional_embeddings()
    # repeat with batch size
    batch_size = pixel_values.shape[0] if pixel_values is not None else image_embeddings.shape[0]
    image_positional_embeddings = image_positional_embeddings.repeat(batch_size, 1, 1, 1)

    vision_attentions = None
    vision_hidden_states = None
    vision_moe_loss = None

    if pixel_values is not None:
        vision_outputs = self.vision_encoder(
            pixel_values,
            output_attentions=output_attentions,
            output_hidden_states=output_hidden_states,
            output_moe_loss=output_moe_loss,
            return_dict=return_dict,
        )
        image_embeddings = vision_outputs[0]

        if output_hidden_states:
            vision_hidden_states = vision_outputs[1]
        if output_attentions:
            # vision_attentions = vision_outputs[-1]
            vision_attentions = vision_outputs[2]
        if output_moe_loss:
            vision_moe_loss = vision_outputs[-1]

    if input_points is not None and input_labels is None:
        input_labels = torch.ones_like(input_points[:, :, :, 0], dtype=torch.int, device=input_points.device)

    if input_points is not None and image_embeddings.shape[0] != input_points.shape[0]:
        raise ValueError(
            "The batch size of the image embeddings and the input points must be the same. ",
            "Got {} and {} respectively.".format(image_embeddings.shape[0], input_points.shape[0]),
            " if you want to pass multiple points for the same image, make sure that you passed ",
            " input_points of shape (batch_size, point_batch_size, num_points_per_image, 3) and ",
            " input_labels of shape (batch_size, point_batch_size, num_points_per_image)",
        )

    sparse_embeddings, dense_embeddings = self.prompt_encoder(
        input_points=input_points,
        input_labels=input_labels,
        input_boxes=input_boxes,
        input_masks=input_masks,
    )

    low_res_masks, iou_predictions, mask_decoder_attentions, class_predictions = self.mask_decoder(
        image_embeddings=image_embeddings,
        image_positional_embeddings=image_positional_embeddings,
        sparse_prompt_embeddings=sparse_embeddings,
        dense_prompt_embeddings=dense_embeddings,
        multimask_output=multimask_output,
        attention_similarity=attention_similarity,
        target_embedding=target_embedding,
        output_attentions=output_attentions,
    )

    if not return_dict:
        output = (
            iou_predictions,
            low_res_masks,
            vision_hidden_states,
            vision_attentions,
            mask_decoder_attentions,
            vision_moe_loss,
        )
        return output

    return (
        SamImageSegmentationOutput(
            iou_scores=iou_predictions,
            pred_masks=low_res_masks,
            vision_hidden_states=vision_hidden_states,
            vision_attentions=vision_attentions,
            mask_decoder_attentions=mask_decoder_attentions,
            vision_moe_loss=vision_moe_loss,
        ),
        ################ New added. Return class predictions as well.
        class_predictions,
        ################
    )


class SAMForSemanticSegmentation(nn.Module):
    """
    Support SAM for semantic segmentation.
    Refer to https://huggingface.co/docs/transformers/main/model_doc/sam
    """

    def __init__(
        self,
        prefix: str,
        checkpoint_name: str,
        num_classes: int = 1,
        pretrained: Optional[bool] = True,
        frozen_layers: Optional[list] = None,
        num_mask_tokens: int = 1,
        image_norm: Optional[str] = None,
    ):
        """
        Load a pretrained Segment Anything Model (SAM).

        Parameters
        ----------
        prefix
            The prefix of the SAMForSemanticSegmentation model.
        checkpoint_name
            Name of the SAM checkpoint.
        num_classes
            The number of classes
        pretrained
            Whether using the pretrained SAM models. If pretrained=True, download the pretrained model.
        frozen_layers
            A list of substrings of frozen layers' names.
        num_mask_tokens
            The number of mask proposals.
        image_norm
            How to normalize an image. We now support:
            - inception
                Normalize image by IMAGENET_INCEPTION_MEAN and IMAGENET_INCEPTION_STD from timm
            - imagenet
                Normalize image by IMAGENET_DEFAULT_MEAN and IMAGENET_DEFAULT_STD from timm
            - clip
                Normalize image by mean (0.48145466, 0.4578275, 0.40821073) and
                std (0.26862954, 0.26130258, 0.27577711), used for CLIP.
        """

        super().__init__()
        self.prefix = prefix
        self.pretrained = pretrained
        self.checkpoint_name = checkpoint_name
        self.num_classes = num_classes
        self.frozen_layers = frozen_layers

        self.device = None
        self.name_to_id = {}

        self._load_checkpoint(checkpoint_name)

        freeze_model_layers(self.model, self.frozen_layers)

        self.image_size = self.model.vision_encoder.image_size
        self.config = self.model.config
        self.image_mean, self.image_std = image_mean_std(image_norm)

        self.model.mask_decoder.num_mask_tokens = num_mask_tokens
        mask_token_data = self.model.mask_decoder.mask_tokens.weight.data[0]
        self.model.mask_decoder.mask_tokens = nn.Embedding(num_mask_tokens, self.model.mask_decoder.hidden_size)
        for i in range(num_mask_tokens):
            self.model.mask_decoder.mask_tokens.weight.data[i] = mask_token_data
        hyper_mlps = self.model.mask_decoder.output_hypernetworks_mlps[0]
        self.model.mask_decoder.output_hypernetworks_mlps = nn.ModuleList([hyper_mlps])
        if num_classes > 1:
            self.model.mask_decoder.num_classes = num_classes
            self.model.mask_decoder.output_classifier_mlps = nn.Linear(
                self.model.mask_decoder.hidden_size, num_classes + 1
            )

            mask_decoder_forward = multi_class_mask_decoder_forward.__get__(
                self.model.mask_decoder, self.model.mask_decoder.__class__
            )
            setattr(self.model.mask_decoder, "forward", mask_decoder_forward)

            sam_model_forward = multi_class_sam_model_forward.__get__(self.model, self.model.__class__)
            setattr(self.model, "forward", sam_model_forward)

        # for Conv-LoRA
        self.output_moe_loss = False

    def _load_checkpoint(self, checkpoint_name):
        if self.pretrained:
            self.model = SamModel.from_pretrained(checkpoint_name)
        else:
            configuration = SamConfig(name_or_path=checkpoint_name)
            self.model = SamModel(configuration)

    def save(self, save_path: str = "./"):
        self.model.save_pretrained(save_path)
        logger.info(f"Model weights for {self.prefix} is saved to {save_path}.")

    @property
    def image_key(self):
        return f"{self.prefix}_{IMAGE}"

    @property
    def image_valid_num_key(self):
        return f"{self.prefix}_{IMAGE_VALID_NUM}"

    @property
    def label_key(self):
        return f"{self.prefix}_{LABEL}"

    @property
    def image_column_prefix(self):
        return f"{self.image_key}_{COLUMN}"

    @property
    def image_feature_dim(self):
        return self.model.num_features

    @property
    def mask_label_key(self):
        return f"{self.prefix}_{MASK_LABEL}"

    @property
    def class_label_key(self):
        return f"{self.prefix}_{CLASS_LABEL}"

    def train(self, mode: bool = True):
        super().train(mode)
        for module in self.modules():
            if isinstance(module, ConvLoRALinear):
                self.output_moe_loss = True
                return self

        return self

    def forward(
        self,
        batch,
    ):
        """
        Parameters
        ----------
        batch
            A dictionary containing the input mini-batch data.
            We need to use the keys with the model prefix to index required data.

        Returns
        -------
            A dictionary with mask predictions.
        """
        # binary
        if self.num_classes == 1:
            rets = self.model(batch[self.image_key], multimask_output=False, output_moe_loss=self.output_moe_loss)
            pred_masks = rets.pred_masks[:, 0, :, :, :]
            pred_masks = F.interpolate(
                pred_masks, (self.image_size, self.image_size), mode="bilinear", align_corners=False
            )
            if self.training:
                rets_dict = {self.prefix: {LOGITS: pred_masks}}
            else:
                rets_dict = {self.prefix: {LOGITS: pred_masks, LABEL: batch[self.label_key]}}
        # multi-class
        else:
            rets = self.model(batch[self.image_key], multimask_output=False, output_moe_loss=self.output_moe_loss)
            rets, class_predictions = rets
            pred_masks = rets.pred_masks[:, 0, :, :, :]
            pred_classes = class_predictions[:, 0, :, :]
            pred_masks = F.interpolate(
                pred_masks, (self.image_size, self.image_size), mode="bilinear", align_corners=False
            )
            if self.training:
                rets_dict = {self.prefix: {LOGITS: pred_masks, CLASS_LOGITS: pred_classes}}
            else:
                rets_dict = {
                    self.prefix: {
                        LOGITS: pred_masks,
                        CLASS_LOGITS: pred_classes,
                        LABEL: batch[self.label_key],
                    }
                }
        if self.output_moe_loss:
            rets_dict[self.prefix].update({MOE_LOSS: rets.vision_moe_loss})
        return rets_dict

    def get_layer_ids(self):
        """
        Assign an id to each layer. Layer ids will be used in layer-wise lr decay.
        Basically, id gradually increases when going from the output end to
        the input end. The layers defined in this class, e.g., head, have id 0.

        In the AutoModel scenario, this function may not always return the correct result.
        Thus, you can use "print(json.dumps(name_to_id, indent=2))" to manually check whether
        the layer ids are reasonable.

        Returns
        -------
        A dictionary mapping the layer names (keys) to their ids (values).
        """
        model_prefix = "model"
        pre_encoder_patterns = (
            "vision_encoder",
            "prompt_encoder",
        )
        post_encoder_patterns = ("mask_decoder",)

        names = [n for n, _ in self.named_parameters()]

        name_to_id, names = assign_layer_ids(
            names=names,
            pre_encoder_patterns=pre_encoder_patterns,
            post_encoder_patterns=post_encoder_patterns,
            model_pre=model_prefix,
        )
        if len(names) > 0:
            logger.debug(f"outer layers are treated as head: {names}")
        for n in names:
            assert n not in name_to_id
            name_to_id[n] = 0

        return name_to_id
